from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
__author__ = 'Stefano Tranquillini <stefano@chino.io>'


class ApiError(Exception):
    code = 500

    def __init__(self, message):
        super(ApiError, self).__init__(message)


class ClientError(Exception):
    def __init__(self, message):
        super(ClientError, self).__init__(message)


class CallError(ApiError):
    def __init__(self, code, message):
        super(CallError, self).__init__(message)
        self.code = code
        if isinstance(message, list):
            self.message = ', '.join(message)
        else:
            self.message = message


class CallFail(CallError):
    def __init__(self, code, message):
        super(CallFail, self).__init__(code, message)


class MethodNotSupported(ClientError):
    def __init__(self):
        super(MethodNotSupported, self).__init__("Method not supported")



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# -*- coding: utf-8 -*-

from datetime import timedelta, datetime

import pytz

"""
Chino.io API Client Docs
~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2015 by Chino SrlS
:license: Apache 2.0, see LICENSE for more details.
"""
import hashlib
import json
import os
import requests

from deprecation import deprecated
from requests.auth import HTTPBasicAuth, AuthBase

from .exceptions import MethodNotSupported, CallError, CallFail, ClientError
from .objects import Repository, ListResult, User, Group, Schema, Document
from .objects import Blob, BlobDetail, UserSchema, Collection, Permission, IDs
from .objects import Application, Consent, DumpFile

import logging.config

__author__ = 'Stefano Tranquillini <stefano@chino.io>'

logger = logging.getLogger('chino.api')


class ChinoAPIBase(object):  # PRAGMA: NO COVER
    """
        Base class, contains the utils methods to call the APIs
    """
    _url = None
    auth = None
    timeout = 30

    def __init__(self, auth, url, timeout, session=True, force_https=True):
        """
        Init the class, auth is ref, so it can be changed and changes
        applies to all the other classes.

        :param auth:
        :param url:
        :return:
        """
        # we have to ensure https, the 302 on http is forwarded
        # to the GET operation even if it's a POST PUT DELETE
        if force_https:
            self._url = url.replace('http://', 'https://')
        else:
            self._url = url
        self.auth = auth
        self.timeout = timeout
        if session:
            self.req = requests.Session()
        else:
            self.req = requests

    # UTILS
    def apicall(self, method, url, params=None, data=None, form=None,
                raw=False):
        method = method.upper()
        url = self._url + url
        if method == 'CHUNK':
            logger.debug("calling %s (PUT) %s p(%s) d(--) " %
                         (method, url, params))
            res = self._apicall_chunk(url, data, **params)
        else:
            logger.debug("calling %s %s p(%s) d(%s) " %
                         (method, url, params, str(data)))
            if method == 'GET':
                res = self._apicall_get(url, params)
            elif method == 'POST':
                res = self._apicall_post(url, data, params, form)
            elif method == 'PUT':
                res = self._apicall_put(url, data)
            elif method == 'DELETE':
                res = self._apicall_delete(url, params)
            elif method == 'PATCH':
                res = self._apicall_patch(url, data)
            else:
                raise MethodNotSupported
        if not raw:
            self.valid_call(res)
            ret = res.json()
            logger.debug("result: %s " % json.dumps(ret))
            data = ret['data']
        else:
            self.valid_call_raw(res)
            data = res
        logger.debug("time: %ss" % res.elapsed.total_seconds())
        logger.debug("-----")
        return data

    def _apicall_chunk(self, url, data, offset, length):
        r = self.req.put(url, data=data, auth=self._get_auth(),
                         headers={'Content-Type': 'application/octet-stream',
                                  'offset': str(offset),
                                  'length': str(length)})
        return r

    def _apicall_put(self, url, data):
        if hasattr(data, 'to_dict()'):
            d = data.to_dict()
        else:
            d = data
        r = self.req.put(url, auth=self._get_auth(),
                         data=json.dumps(d), timeout=self.timeout)
        return r

    def _apicall_patch(self, url, data):
        if hasattr(data, 'to_dict()'):
            d = data.to_dict()
        else:
            d = data
        r = self.req.patch(url, auth=self._get_auth(),
                           data=json.dumps(d), timeout=self.timeout)
        return r

    def _apicall_post(self, url, data, params, form=None):
        if data is None and form is not None:
            r = self.req.post(url, auth=self._get_auth(),
                              params=params, data=form, timeout=self.timeout)
        else:
            if hasattr(data, 'to_dict()'):
                d = data.to_dict()
            else:
                d = data
            r = self.req.post(url, auth=self._get_auth(
            ), params=params, data=json.dumps(d), timeout=self.timeout)
        return r

    def _apicall_get(self, url, params):
        r = self.req.get(url, auth=self._get_auth(),
                         params=params, timeout=self.timeout)
        return r

    def _apicall_delete(self, url, params):
        r = self.req.delete(url, auth=self._get_auth(),
                            params=params, timeout=self.timeout)
        return r

    def _get_auth(self):
        return self.auth.get_auth()

    @staticmethod
    def valid_call(r):
        # logger.debug("%s Response %s ", r.request.url, r.json())
        if r.status_code == requests.codes.ok:
            return True
        else:
            try:
                status = r.json()['result']
            except:
                try:
                    message = r.json()['message']
                except:
                    try:
                        message = r.json()
                    except:
                        message = r.text
                logger.exception(f"Decoding result {r}")
                raise CallError(
                    code=500, message=message)
            if status == 'error':
                raise CallError(code=r.status_code,
                                message=r.json()['message'])
            elif status == 'fail':
                raise CallFail(code=r.status_code, message=r.json()['data'])
            else:
                raise CallError(code=r.status_code, message=r.json())

    @staticmethod
    def valid_call_raw(r):
        # logger.debug("%s Response %s ", r.request.url, r.json())
        if r.status_code == requests.codes.ok:
            return True
        else:
            raise CallError(code=r.status_code, message=r.json())


class ChinoAPIUsers(ChinoAPIBase):
    def __init__(self, *args, **kwargs):
        super(ChinoAPIUsers, self).__init__(*args, **kwargs)

    # TODO: expiration time.

    def code(self, code, redirect_uri, client_id, client_secret):
        # remove auth and save in temp var (in case of problems)

        auth = self.auth
        # self.auth = None
        url = "auth/token/"
        pars = dict(code=code, redirect_uri=redirect_uri, client_id=client_id,
                    client_secret=client_secret,
                    grant_type='authorization_code')
        try:
            self.auth.set_auth_application()
            result = self.apicall('POST', url, form=pars)
            self.auth.refresh_token = result['refresh_token']
            self.auth.bearer_token = result['access_token']
            self.auth.set_auth_user()
            return result
        except Exception as ex:
            # reset auth if things go wrong
            self.auth = auth
            # propagate exception
            raise ex

    def login(self, username, password):
        # remove auth and save in temp var (in case of problems)

        auth = self.auth
        # self.auth = None
        url = "auth/token/"
        pars = dict(username=username, password=password,
                    client_id=self.auth.client_id,
                    client_secret=self.auth.client_secret,
                    grant_type='password')

        try:
            self.auth.set_auth_application()
            result = self.apicall('POST', url, form=pars)
            self.auth.refresh_token = result['refresh_token']
            self.auth.bearer_token = result['access_token']
            self.auth.bearer_exp = datetime.now() + timedelta(
                seconds=int(result['expires_in']))
            self.auth.set_auth_user()
            return result
        except Exception as ex:
            # reset auth if things go wrong
            self.auth = auth
            # propagate exception
            raise ex

    def refresh(self, refresh_token=None):
        # remove auth and save in temp var (in case of problems)
        auth = self.auth
        # self.auth = None
        url = "auth/token/"
        pars = dict(grant_type='refresh_token', client_id=self.auth.client_id,
                    client_secret=self.auth.client_secret,
                    refresh_token=refresh_token if refresh_token else self.auth.refresh_token)

        try:
            self.auth.set_auth_null()
            result = self.apicall('POST', url, form=pars)
            self.auth.refresh_token = result['refresh_token']
            self.auth.bearer_token = result['access_token']
            self.auth.set_auth_user()
            return result
        except Exception as ex:
            # reset auth if things go wrong
            self.auth = auth
            # propagate exception
            raise ex

    def current(self):
        url = "users/me"
        return User(**self.apicall('GET', url)['user'])

    def logout(self):
        url = "auth/revoke_token/"
        auth = self.auth
        pars = dict(client_id=self.auth.client_id,
                    client_secret=self.auth.client_secret,
                    token=self.auth.bearer_token)
        try:
            self.auth.set_auth_null()
            result = self.apicall('POST', url, form=pars)
            return result
        except Exception as ex:
            # reset auth if things go wrong
            self.auth = auth
            # propagate exception
            raise ex

    def list(self, user_schema_id, **pars):
        url = "user_schemas/%s/users" % user_schema_id
        return ListResult(User, self.apicall('GET', url, params=pars))

    def detail(self, user_id):
        url = "users/%s" % user_id
        return User(**self.apicall('GET', url)['user'])

    def create(self, user_schema_id, username, password, attributes=None,
               consistent=False):
        data = dict(username=username, password=password,
                    attributes=attributes)
        url = "user_schemas/%s/users" % user_schema_id
        params = dict()
        if consistent:
            params['consistent'] = 'true'
        return User(
            **self.apicall('POST', url, data=data, params=params)['user'])

    def update(self, user_id, **kwargs):
        url = "users/%s" % user_id
        u_updated = self.apicall('PUT', url, data=kwargs)['user']
        return User(**u_updated)

    def partial_update(self, user_id, consistent=False, **kwargs):
        url = "users/%s" % user_id
        params = dict()
        if consistent:
            params['consistent'] = 'true'
        u_updated = self.apicall('PATCH', url, data=kwargs, params=params)[
            'user']
        return User(**u_updated)

    def delete(self, user_id, force=False, consistent=False):
        url = "users/%s" % user_id
        params = dict()
        if consistent:
            params['consistent'] = 'true'
        if force:
            params['force'] = 'true'

        return self.apicall('DELETE', url, params)

    def search(self, user_schema_id, query, result_type="FULL_CONTENT",
               sort=None, **kwargs):
        url = 'search/users/%s' % user_schema_id
        if not sort:
            sort = []
        data = dict(result_type=result_type, query=query)
        if sort:
            data['sort'] = sort
        if result_type == "COUNT":
            return self.apicall('POST', url, data=data, params=kwargs)['count']
        elif result_type == "EXISTS" or result_type == "USERNAME_EXISTS":
            return self.apicall('POST', url, data=data, params=kwargs)[
                'exists']
        else:
            return ListResult(User, self.apicall('POST', url, data=data,
                                                 params=kwargs))


class ChinoAPIGroups(ChinoAPIBase):
    def __init__(self, *args, **kwargs):
        super(ChinoAPIGroups, self).__init__(*args, **kwargs)

    def list(self, **pars):
        """
        Gets the list of repository

        :param: usual for a list ``offset``, ``limit``. Use ``name`` to
                filter by group name
        """
        url = "groups"
        return ListResult(Group, self.apicall('GET', url, params=pars))

    def detail(self, group_id):
        url = "groups/%s" % group_id
        return Group(**self.apicall('GET', url)['group'])

    def create(self, groupname, attributes=None):
        data = dict(group_name=groupname, attributes=attributes)
        url = "groups"
        return Group(**self.apicall('POST', url, data=data)['group'])

    def update(self, group_id, **kwargs):
        url = "groups/%s" % group_id
        return self.apicall('PUT', url, data=kwargs)['group']

    def delete(self, group_id, force=False):
        url = "groups/%s" % group_id
        if force:
            params = dict(force='true')
        else:
            params = None
        return self.apicall('DELETE', url, params)

    def add_user(self, group_id, user_id):
        url = "groups/%s/users/%s" % (group_id, user_id)
        return self.apicall('POST', url)

    def del_user(self, group_id, user_id):
        url = "groups/%s/users/%s" % (group_id, user_id)
        return self.apicall('DELETE', url)

    def list_users(self, group_id, **pars):
        url = "groups/%s/users" % group_id
        return ListResult(User, self.apicall('GET', url, params=pars))


class ChinoAPIPermissions(ChinoAPIBase):
    def __init__(self, *args, **kwargs):
        super(ChinoAPIPermissions, self).__init__(*args, **kwargs)

    def resources(self, action, resource_type, subject_type, subject_id,
                  manage=None, authorize=None):
        url = "perms/%s/%s/%s/%s" % (action,
                                     resource_type, subject_type, subject_id)
        data = dict()
        if manage:
            data['manage'] = manage
        if authorize:
            data['authorize'] = authorize
        return self.apicall('POST', url, data=data)

    def resource(self, action, resource_type, resource_id, subject_type,
                 subject_id,
                 manage=None, authorize=None):
        url = "perms/%s/%s/%s/%s/%s" % (action, resource_type,
                                        resource_id, subject_type, subject_id)
        data = dict()
        if manage:
            data['manage'] = manage
        if authorize:
            data['authorize'] = authorize
        return self.apicall('POST', url, data=data)

    def resource_children(self, action, resource_type, resource_id,
                          resource_child_type, subject_type, subject_id,
                          manage=None, authorize=None, created_document=None):
        url = "perms/%s/%s/%s/%s/%s/%s" % (
            action, resource_type, resource_id, resource_child_type,
            subject_type, subject_id)
        data = dict()
        if manage:
            data['manage'] = manage
        if authorize:
            data['authorize'] = authorize
        if created_document:
            data['created_document'] = created_document
        return self.apicall('POST', url, data=data)

    def read_perms(self):
        url = "perms"
        res = self.apicall('GET', url)
        return [Permission(**p) for p in res['permissions']]

    def read_perms_document(self, document_id):
        url = "perms/documents/%s" % document_id
        return [Permission(**p) for p in
                self.apicall('GET', url)['permissions']]

    def read_perms_user(self, user_id):
        url = "perms/users/%s" % user_id
        return [Permission(**p) for p in
                self.apicall('GET', url)['permissions']]

    def read_perms_group(self, group_id):
        url = "perms/groups/%s" % group_id
        return [Permission(**p) for p in
                self.apicall('GET', url)['permissions']]


class ChinoAPIRepositories(ChinoAPIBase):
    def __init__(self, *args, **kwargs):
        super(ChinoAPIRepositories, self).__init__(*args, **kwargs)

    def list(self, **pars):
        """
        Gets the list of repository

        :param: usual for a list ``offset``, ``limit``. Use ``descr`` to
                filter by description
        :return: dict containing ``count``,``total_count``,``limit``,``offset``
            and the list of items
        inside a property with its name (e.g., ``documents``)
        """
        url = "repositories"
        return ListResult(Repository, self.apicall('GET', url, params=pars))

    def detail(self, repository_id):
        """
        Gets the details of repository.

        :param repository_id: (id) the id of the repository
        :return: (dict) the repository.
        """
        url = "repositories/%s" % repository_id
        return Repository(**self.apicall('GET', url)['repository'])

    def create(self, description):
        """
        Creates a a repository.

        :param description: (str) the name of the repository
        :return: (dict) the repository.
        """
        data = dict(description=description)
        url = "repositories"
        return Repository(**self.apicall('POST', url, data=data)['repository'])

    def update(self, repository_id, **kwargs):
        """
        Update a a repository.

        :param repository_id: (id) the id of the repository
        :return: (dict) the repository.
        """
        url = "repositories/%s" % repository_id
        return Repository(
            **self.apicall('PUT', url, data=kwargs)['repository'])

    def delete(self, repository_id, force=False, all_content=False):
        """
        Creates a a repository.

        :param repository_id: (id) the id of the repository
        :param force: to force a deletion
        :param all_content: to delete in cascade
        :return: None
        """
        url = "repositories/%s" % repository_id
        params = dict()
        if force:
            params['force'] = 'true'
        if all_content:
            params['all_content'] = 'true'
        return self.apicall('DELETE', url, params)


class ChinoAPISchemas(ChinoAPIBase):
    def __init__(self, *args, **kwargs):
        super(ChinoAPISchemas, self).__init__(*args, **kwargs)

    def list(self, repository_id, **pars):
        """
        Gets the list of documents by schema

        :param repository_id: (id) the id of the repository
        :param: usual for a list ``offset``, ``limit``. Use ``descr`` to
                filter by description
        :return: dict containing ``count``,``total_count``,``limit``,
            ``offset``,``repositories``
        """
        url = "repositories/%s/schemas" % repository_id
        return ListResult(Schema, self.apicall('GET', url, params=pars))

    def create(self, repository, description, fields):
        """
        Creates a schema in a repository.

        :param repository: (id) the repository in which the schema is created
        :param description: (str) the name of the schema
        :param fields: list(dict) the list of fields
        :return: (dict) the schema.
        """
        data = dict(description=description, structure=dict(fields=fields))
        url = "repositories/%s/schemas" % repository
        return Schema(**self.apicall('POST', url, data=data)['schema'])

    def detail(self, schema_id):
        """
        Details of a schema in a repository.

        :param schema_id: (id) of the schema
        :return: (dict) the schema.
        """
        url = "schemas/%s" % schema_id
        return Schema(**self.apicall('GET', url)['schema'])

    def update(self, schema_id, **kwargs):
        url = "schemas/%s" % schema_id
        return Schema(**self.apicall('PUT', url, data=kwargs)['schema'])

    def delete(self, schema_id, force=False, all_content=False):
        url = "schemas/%s" % schema_id
        params = dict()
        if force:
            params['force'] = 'true'
        if all_content:
            params['all_content'] = 'true'
        return self.apicall('DELETE', url, params)

    def convert_datatypes_submit(self, schema_id, fields=None, defaults=None):
        """
        Submit a new convert datatypes job.

        :param schema_id: (id) of the schema
        :param fields: 'fields' (dict) field name: new datatype
        :param defaults: (dict) field name: new default
        :return: message
        """
        url = "schemas/convert/%s" % schema_id
        data = {}
        if fields:
            data['fields'] = fields
        if defaults:
            data['defaults'] = defaults

        return self.apicall('POST', url, data=data)

    def convert_datatypes_status(self, schema_id):
        """
        Read the current status of the datatype conversion job

        :param schema_id: (id) of the schema
        :return: dict
        """
        url = "schemas/convert/%s/status" % schema_id
        return self.apicall('GET', url)

    def dump_submit(self, schema_id, query, sort=None, fields=None,
                    format='json'):
        """
        Submit a new dump schema job.

        :param schema_id: (id) of the schema
        :param query: (dict) a query to limit the documents to dump
        :param sort: (dict) sort info
        :param fields: (list) a list to limit the fields to dump
        :param format: (str) the format of the ouput file (json, csv)
        :return: dict with submit recap and the job id (dump_id)
        """
        url = "schemas/dump/%s" % schema_id
        data = {'query': query, 'format': format}
        if sort:
            data['sort'] = sort
        if fields:
            data['fields'] = fields
        return self.apicall('POST', url, data=data)

    def dump_status(self, dump_id):
        """
        Read the status of the dump job.

        :param dump_id: (str) the id of the dump job (uuid)
        :return: (dict) info about the status of the job
        """
        url = "schemas/dump/%s/status" % dump_id
        return self.apicall('GET', url)

    def dump_download(self, dump_id):
        """
        Download the file with the dump.

        :param dump_id: (str) the id of the dump job (uuid)
        :retrieve_status: (bool) if True fill the DumpFile obj with
                                 status info (default)
        :return: (stream) the stream of the dump file
        """
        status = self.dump_status(dump_id)
        if not status['status'] == 'COMPLETED':
            raise CallError(400, "Download not ready")

        url = "schemas/dump/%s/download" % dump_id
        dump_resp = self.apicall('GET', url, raw=True)
        if dump_resp.status_code == 200:
            return DumpFile(dump_id, dump_resp.content, status_info=status)
        else:
            return dump_resp


class ChinoAPIDocuments(ChinoAPIBase):
    def __init__(self, *args, **kwargs):
        super(ChinoAPIDocuments, self).__init__(*args, **kwargs)

    def list(self, schema_id, full_document=False, **pars):
        url = "schemas/%s/documents" % schema_id
        if full_document:
            pars['full_document'] = 'true'

        # validate is_active and convert bool to the required bool string
        is_active = pars.get('is_active', None)
        if is_active is not None:
            if is_active not in (True, False):
                raise TypeError("is_active must be bool")
            pars['is_active'] = str(is_active).lower()

        # validate datetime params and convert to the required datetime string
        datetime_params = [
            'insert_date', 'insert_date__gt', 'insert_date__lt',
            'last_update', 'last_update__gt', 'last_update__lt',
        ]
        for param in datetime_params:
            d = pars.get(param, None)
            if d is None:
                continue

            if not isinstance(d, datetime):
                raise TypeError("{} must be datetime".format(param))

            # if it's not a naive datetime (has timezone info), make it UTC
            if d.tzinfo is not None and d.tzinfo.utcoffset(d) is not None:
                d = d.astimezone(pytz.utc)

            pars[param] = d.strftime("%Y-%m-%dT%H:%M:%S")

        return ListResult(Document, self.apicall('GET', url, params=pars))

    def create(self, schema_id, content, consistent=False):
        data = dict(content=content)
        url = "schemas/%s/documents" % schema_id
        params = dict()
        if consistent:
            params['consistent'] = 'true'
        return Document(
            **self.apicall('POST', url, data=data, params=params)['document'])

    def detail(self, document_id):
        url = "documents/%s" % document_id
        return Document(**self.apicall('GET', url)['document'])

    def update(self, document_id, **kwargs):
        # if partial=True is passed, use PATCH instead of PUT
        partial = kwargs.pop('partial', False)
        method = 'PATCH' if partial else 'PUT'

        # data = dict(content=content)
        url = "documents/%s" % document_id
        return Document(**self.apicall(method, url, data=kwargs)['document'])

    def delete(self, document_id, force=False, consistent=False):
        url = "documents/%s" % document_id
        params = dict()
        if force:
            params['force'] = 'true'
        if consistent:
            params['consistent'] = 'true'
        return self.apicall('DELETE', url, params)

    def search(self, schema_id, query, result_type="FULL_CONTENT", sort=None,
               **kwargs):
        url = 'search/documents/%s' % schema_id
        if not sort:
            sort = []
        data = dict(result_type=result_type, query=query)
        if sort:
            data['sort'] = sort
        if result_type == "COUNT":
            return self.apicall('POST', url, data=data, params=kwargs)['count']
        elif result_type == "EXISTS":
            return bool(
                self.apicall('POST', url, data=data, params=kwargs)['exists'])
        elif result_type == "ONLY_ID":
            return ListResult(IDs, self.apicall('POST', url, data=data,
                                                params=kwargs))
        else:
            return ListResult(Document, self.apicall('POST', url, data=data,
                                                     params=kwargs))


class ChinoAPIBlobs(ChinoAPIBase):
    def __init__(self, *args, **kwargs):
        super(ChinoAPIBlobs, self).__init__(*args, **kwargs)

    def _read_bytes_from_file(self, filename, chunksize=32 * 1024):
        """
        for not enctypted
        :param filename:
        :param chunksize:
        :return:
        """
        with open(filename, "rb") as f:
            while True:
                chunk = f.read(chunksize)
                if chunk:
                    yield chunk
                else:
                    break
            f.close()

    def send(self, document_id, blob_field_name, file_path,
             chunk_size=12 * 1024):
        if not os.path.exists(file_path):
            raise ClientError("File not found")
        # start the blob
        blob_data = self.start(document_id, blob_field_name,
                               os.path.basename(file_path))
        # get the id and initial offset
        upload_id = blob_data['upload_id']
        # open the file and start reading it
        logger.debug("file size %s", os.path.getsize(file_path))
        sha1 = hashlib.sha1()
        offset = 0
        for chunk in self._read_bytes_from_file(file_path,
                                                chunksize=chunk_size):
            l_chunk = len(chunk)
            # if we have read enough bytes
            logger.debug("chunk %s " % l_chunk)
            # send a chunk

            self.chunk(upload_id, chunk, length=l_chunk, offset=offset)
            # update the hash
            sha1.update(chunk)
            offset += l_chunk
        # commit and check if everything was fine
        commit = self.commit(upload_id)
        if sha1.hexdigest() != commit['sha1']:
            raise CallFail(500, 'The file was not uploaded correctly')
        return BlobDetail(**commit)

    def start(self, document_id, field, field_name):
        url = 'blobs'
        data = dict(document_id=document_id, field=field, file_name=field_name)
        return self.apicall('POST', url, data=data)['blob']

    def chunk(self, upload_id, data, length, offset):
        # here we use directly request library to implement the binary
        url = 'blobs/%s' % upload_id
        # url = "http://httpbin.org/put"
        return self.apicall('CHUNK', url, data=data,
                            params=dict(length=length, offset=offset))['blob']

    def commit(self, upload_id):
        url = 'blobs/commit'
        data = dict(upload_id=upload_id)
        return self.apicall('POST', url, data=data)['blob']

    def detail(self, blob_id):
        # NOTE: this calls directly the function. needed to get the headers
        url = 'blobs/%s' % blob_id
        # this is different
        res = self.apicall('GET', url, raw=True)
        fname = res.headers['Content-Disposition'].split(';')[1].split('=')[1]
        return Blob(filename=fname, content=res.content, blob_id=blob_id)

    def delete(self, blob_id):
        url = 'blobs/%s' % blob_id
        return self.apicall('DELETE', url)

    def generate(self, blob_id, one_time=None, duration=None):
        url = 'blobs/{}/generate'.format(blob_id)
        data = dict()
        if one_time:
            # even if false is not sent, but false is the default
            data['one_time'] = one_time
        if duration:
            data['duration'] = duration
        return self.apicall('POST', url, data=data)

    def detail_token(self, blob_id, token, filename, mimetype=None):
        url = 'blobs/url/%s?token=%s' % (blob_id, token)
        if mimetype:
            url += "&mimetype=%s" % mimetype
        # this is different
        res = self.apicall('GET', url, raw=True)
        # fname = res.headers['Content-Disposition'].split(';')[1].split('=')[1]
        return Blob(filename=filename, content=res.content, blob_id=blob_id)


class ChinoAPISearches(ChinoAPIBase):
    def __init__(self, *args, **kwargs):
        super(ChinoAPISearches, self).__init__(*args, **kwargs)

    @deprecated(deprecated_in="2.4.0", removed_in="3.0.0",
                details="Use ChinoAPIDocuments.search instead")
    def search(self, schema_id, result_type="FULL_CONTENT", filter_type="and",
               sort=None, filters=None, **kwargs):
        return self.documents(schema_id, result_type, filter_type, sort,
                              filters, **kwargs)

    @deprecated(deprecated_in="2.4.0", removed_in="3.0.0",
                details="Use ChinoAPIDocuments.search instead")
    def documents(self, schema_id, result_type="FULL_CONTENT",
                  filter_type="and", sort=None, filters=None, **kwargs):
        url = 'search/documents/%s' % schema_id
        if not sort:
            sort = []
        if not filters:
            filters = []
        data = dict(result_type=result_type,
                    filter_type=filter_type, filter=filters)
        if sort:
            data['sort'] = sort
        if result_type == "COUNT":
            return self.apicall('POST', url, data=data, params=kwargs)['count']
        elif result_type == "EXISTS":
            return bool(
                self.apicall('POST', url, data=data, params=kwargs)['exists'])
        elif result_type == "ONLY_ID":
            return ListResult(IDs, self.apicall('POST', url, data=data,
                                                params=kwargs))
        else:
            return ListResult(Document, self.apicall('POST', url, data=data,
                                                     params=kwargs))

    @deprecated(deprecated_in="2.4.0", removed_in="3.0.0",
                details="Use ChinoAPIDocuments.search instead")
    def documents_complex(self, schema_id, query, result_type="FULL_CONTENT",
                          sort=None,
                          **kwargs):
        url = 'search/documents/%s' % schema_id
        if not sort:
            sort = []
        data = dict(result_type=result_type,
                    query=query)
        if sort:
            data['sort'] = sort
        if result_type == "COUNT":
            return self.apicall('POST', url, data=data, params=kwargs)['count']
        elif result_type == "EXISTS":
            return bool(
                self.apicall('POST', url, data=data, params=kwargs)['exists'])
        elif result_type == "ONLY_ID":
            return ListResult(IDs, self.apicall('POST', url, data=data,
                                                params=kwargs))
        else:
            return ListResult(Document, self.apicall('POST', url, data=data,
                                                     params=kwargs))

    @deprecated(deprecated_in="2.4.0", removed_in="3.0.0",
                details="Use ChinoAPIUsers.search instead")
    def users(self, user_schema_id, result_type="FULL_CONTENT",
              filter_type="and", sort=None, filters=None, **kwargs):
        url = 'search/users/%s' % user_schema_id
        if not sort:
            sort = []
        if not filters:
            filters = []
        data = dict(result_type=result_type,
                    filter_type=filter_type, filter=filters)
        if sort:
            data['sort'] = sort
        if result_type == "COUNT":
            return self.apicall('POST', url, data=data, params=kwargs)['count']
        elif result_type == "EXISTS" or result_type == "USERNAME_EXISTS":
            return bool(
                self.apicall('POST', url, data=data, params=kwargs)['exists'])
        else:
            return ListResult(User, self.apicall('POST', url, data=data,
                                                 params=kwargs))

    @deprecated(deprecated_in="2.4.0", removed_in="3.0.0",
                details="Use ChinoAPIUsers.search instead")
    def users_complex(self, user_schema_id, query, result_type="FULL_CONTENT",
                      sort=None, **kwargs):
        url = 'search/users/%s' % user_schema_id
        if not sort:
            sort = []
        data = dict(result_type=result_type, query=query)
        if sort:
            data['sort'] = sort
        if result_type == "COUNT":
            return self.apicall('POST', url, data=data, params=kwargs)['count']
        elif result_type == "EXISTS" or result_type == "USERNAME_EXISTS":
            return self.apicall('POST', url, data=data, params=kwargs)[
                'exists']
        else:
            return ListResult(User, self.apicall('POST', url, data=data,
                                                 params=kwargs))


class ChinoAuth(object):
    customer_id = None
    customer_key = None
    client_id = None
    client_secret = None
    bearer_token = None
    refresh_token = None
    bearer_exp = None
    __auth = None

    def __init__(self, customer_id, customer_key=None, bearer_token=None,
                 client_id=None, client_secret=None,
                 refresh_token=None, bearer_exp=None):
        """
        Init the class

        :param customer_id: mandatory
        :param customer_key: optional, if specified the auth is set as
                chino customer (admin)
        :param bearer_token: optional, if specified the auth is as user
        :return: the class
        """
        self.customer_id = customer_id
        self.customer_key = customer_key
        self.bearer_token = bearer_token
        self.client_id = client_id
        self.client_secret = client_secret
        self.refresh_token = refresh_token
        self.bearer_exp = bearer_exp

        if customer_key:
            # if customer_key is set, then set auth as that
            self.set_auth_admin()
        elif bearer_token:
            # if access_token is set, then use it as customer
            self.set_auth_user()
        elif client_id:
            self.set_auth_application()

    def set_auth_admin(self):
        self.__auth = HTTPBasicAuth(self.customer_id, self.customer_key)

    def set_auth_user(self):
        self.__auth = HTTPBearerAuth(self.bearer_token)

    def set_auth_null(self):
        self.__auth = None

    def set_auth_application(self):
        if self.client_secret:
            self.__auth = HTTPBasicAuth(self.client_id, self.client_secret)
        else:
            self.set_auth_null()

    def get_auth(self):
        return self.__auth


class HTTPBearerAuth(AuthBase):
    """Attaches HTTP Basic Authentication to the given Request object."""

    def __init__(self, bearer_token):
        self.bearer_token = bearer_token

    def __call__(self, r):
        r.headers['Authorization'] = 'Bearer %s' % self.bearer_token
        return r


class ChinoAPIUserSchemas(ChinoAPIBase):
    def __init__(self, *args, **kwargs):
        super(ChinoAPIUserSchemas, self).__init__(*args, **kwargs)

    def list(self, **pars):
        """
        Gets the list of UserSchemas

        :param: usual for a list ``offset``, ``limit``. Use ``descr`` to
                filter by description
        :return: dict containing ``count``,``total_count``,``limit``,``offset``
            ``repositories``
        """
        url = "user_schemas"
        return ListResult(UserSchema, self.apicall('GET', url, params=pars))

    def create(self, description, fields):
        """
        Creates a UserSchema

        :param description: (str) the name of the schema
        :param fields: list(dict) the list of fields
        :return: (dict) the UserSchema.
        """
        data = dict(description=description, structure=dict(fields=fields))
        url = "user_schemas"
        return UserSchema(
            **self.apicall('POST', url, data=data)['user_schema'])

    def detail(self, user_schema_id):
        """
        Details of a UserSchema

        :param user_schema_id: (id) of the schema
        :return: (dict) the schema.
        """
        url = "user_schemas/%s" % user_schema_id
        return UserSchema(**self.apicall('GET', url)['user_schema'])

    def update(self, user_schema_id, **kwargs):
        """
        Updates a UserSchema
        :param user_schema_id:
        :param kwargs:
        :return:
        """
        url = "user_schemas/%s" % user_schema_id
        return UserSchema(
            **self.apicall('PUT', url, data=kwargs)['user_schema'])

    def delete(self, user_schema_id, force=False):
        url = "user_schemas/%s" % user_schema_id
        if force:
            params = dict(force='true')
        else:
            params = None
        return self.apicall('DELETE', url, params)

    def convert_datatypes(self, user_schema_id, fields=None, defaults=None):
        """
        Submit a new convert datatypes job.

        :param user_schema_id: (id) of the user schema
        :param fields: 'fields' (dict) field name: new datatype
        :param defaults: (dict) field name: new default
        :return: message
        """
        url = "user_schemas/convert/%s" % user_schema_id
        data = {}
        if fields:
            data['fields'] = fields
        if defaults:
            data['defaults'] = defaults

        return self.apicall('POST', url, data=data)

    def convert_datatypes_status(self, user_schema_id):
        """
        Read the current status of the datatype conversion job

        :param user_schema_id: (id) of the user schema
        :return: dict
        """
        url = "user_schemas/convert/%s/status" % user_schema_id
        return self.apicall('GET', url)


class ChinoAPICollections(ChinoAPIBase):
    def __init__(self, *args, **kwargs):
        super(ChinoAPICollections, self).__init__(*args, **kwargs)

    def list(self, **pars):
        """
        Gets the list of Collections

        :param: usual for a list ``offset``, ``limit``. Use ``document_id``
                to list only the collections with a specific document
        :return: dict containing ``count``,``total_count``,``limit``,
                ``offset``,``repositories``
        """
        # if you specify 'document_id' as kwarg:
        #    collections = client.collections.list(document_id=<doc_id>)
        # use the url with the filter, otherwise the base LIST url
        if 'document_id' in pars.keys():
            url = "collections/document/%s" % pars.pop('document_id')
        else:
            url = "collections"

        return ListResult(Collection, self.apicall('GET', url, params=pars))

    def create(self, name):
        """
        Creates a Collection

        :param name: (str) the name of the collection
        :return: (dict) the UserSchema.
        """
        data = dict(name=name)
        url = "collections"
        return Collection(**self.apicall('POST', url, data=data)['collection'])

    def detail(self, collection_id):
        """
        Details of a Collection

        :param collection_id: (id) of the Collection
        :return: (dict) the Collection.
        """
        url = "collections/%s" % collection_id
        return Collection(**self.apicall('GET', url)['collection'])

    def update(self, collection_id, **kwargs):
        url = "collections/%s" % collection_id
        return Collection(
            **self.apicall('PUT', url, data=kwargs)['collection'])

    def delete(self, collection_id, force=False):
        url = "collections/%s" % collection_id
        if force:
            params = dict(force='true')
        else:
            params = None
        return self.apicall('DELETE', url, params)

    def list_documents(self, collection_id, **pars):
        url = "collections/%s/documents" % collection_id
        return ListResult(Document, self.apicall('GET', url, params=pars))

    def add_document(self, collection_id, document_id):
        url = "collections/%s/documents/%s" % (collection_id, document_id)
        return self.apicall('POST', url)

    def rm_document(self, collection_id, document_id):
        url = "collections/%s/documents/%s" % (collection_id, document_id)
        return self.apicall('DELETE', url)

    def search(self, name, contains=False, **pars):
        url = "collections/search"
        data = dict(name=name, contains=contains)
        return ListResult(Collection,
                          self.apicall('POST', url, params=pars, data=data))


class ChinoAPIApplication(ChinoAPIBase):
    def __init__(self, *args, **kwargs):
        super(ChinoAPIApplication, self).__init__(*args, **kwargs)

    def list(self, **pars):
        """
        Gets the list of Application

        :param: usual for a list ``offset``, ``limit``
        :return: dict containing ``count``,``total_count``,``limit``,``offset``
                ,``repositories``
        """
        url = "auth/applications"
        return ListResult(Application, self.apicall('GET', url, params=pars))

    def create(self, name, grant_type='password', redirect_url='',
               client_type='confidential'):
        """
        Creates a Application.


        :param name: (str) the name of the Application
        :param grant_type: the grant_type
        :param redirect_url: the redirect url (if grant_type allows it)
        :param client_type: optional parameter client_type can be either
                'public' or
                'confidential'."confidential" is the default value.
        :return: (dict) the Application.
        """
        data = dict(name=name, grant_type=grant_type,
                    redirect_url=redirect_url, client_type=client_type)
        url = "auth/applications"
        return Application(
            **self.apicall('POST', url, data=data)[Application.__str_name__])

    def detail(self, application_id):
        """
        Details of a Application

        :param application_id: (id) of the Application
        :return: (dict) the Application.
        """
        url = "auth/applications/%s" % application_id
        return Application(
            **self.apicall('GET', url)[Application.__str_name__])

    def update(self, application_id, **kwargs):
        url = "auth/applications/%s" % application_id
        return Application(
            **self.apicall('PUT', url, data=kwargs)[Application.__str_name__])

    def delete(self, application_id, force=False):
        url = "auth/applications/%s" % application_id
        params = dict()
        if force:
            params['force'] = 'true'

        return self.apicall('DELETE', url, params)


class ChinoAPIConsents(ChinoAPIBase):
    def __init__(self, *args, **kwargs):
        super(ChinoAPIConsents, self).__init__(*args, **kwargs)

    @deprecated(deprecated_in="2.4.0", removed_in="3.0.0",
                details="Deprecated in favor of the new Consenta API: "
                        "https://docs.chino.io/consent/consentame/docs/v1")
    def list(self, user_id=None, **pars):
        """
        Gets the list of all available consents.

        :param user_id: optional. If specified, results are filtered by user
        :param pars: dict with fields ``offset``, ``limit``
                (to navigate through paged results of a list)
        :return: dict containing ``count``,``total_count``,``limit``,
                ``offset``,``repositories``
        """
        url = "consents"
        url_params = dict(pars)
        if user_id:
            url_params['user_id'] = user_id
        return ListResult(Consent, self.apicall("GET", url, params=url_params))

    @deprecated(deprecated_in="2.4.0", removed_in="3.0.0",
                details="Deprecated in favor of the new Consenta API: "
                        "https://docs.chino.io/consent/consentame/docs/v1")
    def create(self, user_id, details, data_controller, purposes):
        """
        Creates and returns a new Consent object.

        :param user_id: string that identifies uniquely a user.
                    Can be a chino.objects.User user_id attribute
                    or any other valid string
                    (i.e. an email address)
        :param details: a dict that contains all the following fields:
                    ``description``, ``policy_url``,
                    ``policy_version`` and ``collection_mode``
        :param data_controller: a dict that must contain all the following
                fields: ``company``, ``contact``, ``address``, ``email``,
                ``VAT`` and ``on_behalf``
        :param purposes: a list that contains one or more dicts.
                The dicts must contain all the following fields:
                ``authorized``, ``purpose`` and ``description``
        :return: the created Consent object
        """
        consent_obj = dict(details)
        consent_obj['user_id'] = user_id
        consent_obj['data_controller'] = data_controller
        consent_obj['purposes'] = purposes

        url = "consents"

        return Consent(
            **self.apicall("POST", url, data=consent_obj)['consent'])

    @deprecated(deprecated_in="2.4.0", removed_in="3.0.0",
                details="Deprecated in favor of the new Consenta API: "
                        "https://docs.chino.io/consent/consentame/docs/v1")
    def detail(self, consent_id):
        """
        Gets the Consent object with the specified ID

        :param consent_id: the Chino.io API id of this consent object
        :return: a Consent object with the specified consent_id
        """
        url = "consents/%s" % consent_id
        return Consent(**self.apicall("GET", url=url)['consent'])

    @deprecated(deprecated_in="2.4.0", removed_in="3.0.0",
                details="Deprecated in favor of the new Consenta API: "
                        "https://docs.chino.io/consent/consentame/docs/v1")
    def history(self, consent_id, **pars):
        """
        Get an history of the Consent object with the provided ID

        :param consent_id: the Chino.io API id of this consent object
        :param pars: dict with fields ``offset``, ``limit``
                to navigate through paged results of a list
        :return: a ListResult with the history of the Consent object
                with the specified consent_id
        """
        url = "consents/%s/history" % consent_id
        return ListResult(Consent, self.apicall("GET", url, params=pars))

    @deprecated(deprecated_in="2.4.0", removed_in="3.0.0",
                details="Deprecated in favor of the new Consenta API: "
                        "https://docs.chino.io/consent/consentame/docs/v1")
    def update(self, consent_id, user_id, details, data_controller, purposes):
        """
        Creates a new Consent with the specified informations and withdraws
        the old one which has the same ID.

        :param consent_id: the Chino.io API id of this consent object
        :param user_id: string that identifies the user.
        :param details: a dict that contains all the following fields:
                ``description``, ``policy_url``,
                ``policy_version`` and ``collection_mode``
        :param data_controller: a dict that must contain all the following
                fields: ``company``, ``contact``, ``address``,
                `email``, ``VAT`` and ``on_behalf``
        :param purposes: a list that contains one or more dicts.
                The dicts must contain all the following fields:
                ``authorized``, ``purpose`` and ``description``
        :return: the created Consent object
        """
        updated_consent = dict(details)
        updated_consent['user_id'] = user_id
        updated_consent['data_controller'] = data_controller
        updated_consent['purposes'] = purposes

        url = "consents/%s" % consent_id
        return Consent(
            **self.apicall("PUT", data=updated_consent, url=url)['consent'])

    @deprecated(deprecated_in="2.4.0", removed_in="3.0.0",
                details="Deprecated in favor of the new Consenta API: "
                        "https://docs.chino.io/consent/consentame/docs/v1")
    def withdraw(self, consent_id):
        """
        Withdraws the Consent which has the specified ID. Withdrawn consents
        can not be further updated, but remain in the database for further
        reference.

        :params consent_id: the Chino.io API id of this consent object
        :return: the object that is returned by the API call.
        """
        url = "consents/%s" % consent_id
        return self.apicall('DELETE', url)

    @deprecated(deprecated_in="2.4.0", removed_in="3.0.0",
                details="Deprecated in favor of the new Consenta API: "
                        "https://docs.chino.io/consent/consentame/docs/v1")
    def delete(self, consent_id):
        """
        only works with test API. Deletes Consent from server
        """
        url = "consents/%s" % consent_id
        return self.apicall("DELETE", url, params={'force': 'true'})


class ChinoAPIClient(object):
    """
    ChinoAPI the client class
    """
    final_url = ""
    users = groups = permissions = repositories = schemas = documents = blobs = searches = None

    def __init__(self, customer_id, customer_key=None, bearer_token=None,
                 client_id=None, client_secret=None,
                 version='v1', url='https://api.chino.io/', timeout=30,
                 session=True, force_https=True):
        """
        Init the class

        :param customer_id: mandatory
        :param customer_key: optional, if specified the auth is set as chino
                        customer (admin)
        :param bearer_token: optional, if specified the auth is as user
        :param version: default is `v1`, change if you know what to do
        :param url: the url, this should be changed only for testing
        :return: the class
        """

        # smarter way to add slash?
        if not url.endswith('/'):
            url += '/'
        if not version.endswith('/'):
            version += '/'
        final_url = url + version
        self.final_url = final_url
        auth = ChinoAuth(customer_id, customer_key,
                         bearer_token, client_id, client_secret)
        self.auth = auth
        self.users = ChinoAPIUsers(
            auth, final_url, timeout=timeout, session=session,
            force_https=force_https)
        self.applications = ChinoAPIApplication(
            auth, final_url, timeout=timeout, session=session,
            force_https=force_https)
        self.groups = ChinoAPIGroups(
            auth, final_url, timeout=timeout, session=session,
            force_https=force_https)
        self.permissions = ChinoAPIPermissions(
            auth, final_url, timeout=timeout, session=session,
            force_https=force_https)
        self.repositories = ChinoAPIRepositories(
            auth, final_url, timeout=timeout, session=session,
            force_https=force_https)
        self.schemas = ChinoAPISchemas(
            auth, final_url, timeout=timeout, session=session,
            force_https=force_https)
        self.user_schemas = ChinoAPIUserSchemas(
            auth, final_url, timeout=timeout, session=session,
            force_https=force_https)
        self.collections = ChinoAPICollections(
            auth, final_url, timeout=timeout, session=session,
            force_https=force_https)
        self.documents = ChinoAPIDocuments(
            auth, final_url, timeout=timeout, session=session,
            force_https=force_https)
        self.blobs = ChinoAPIBlobs(
            auth, final_url, timeout=timeout, session=session,
            force_https=force_https)
        self.searches = ChinoAPISearches(
            auth, final_url, timeout=timeout, session=session,
            force_https=force_https)
        self.consents = ChinoAPIConsents(
            auth, final_url, timeout=timeout, session=session,
            force_https=force_https)



import hashlib
import logging
import logging.config
import os
import time
import unittest

import pytz

from chino.api import ChinoAPIClient
from chino.exceptions import CallError
from chino.objects import _DictContent, _Field
from datetime import datetime, timedelta

from . import cfg


__author__ = 'Stefano Tranquillini <stefano@chino.io>'


# logging.config.fileConfig(path.join(path.dirname(__file__), 'logging.conf'))


class BaseChinoTest(unittest.TestCase):
    def setUp(self):
        self.chino = ChinoAPIClient(customer_id=cfg.customer_id,
                                    customer_key=cfg.customer_key,
                                    url=cfg.url, client_id=cfg.client_id,
                                    client_secret=cfg.client_secret,
                                    timeout=3600, force_https=False)
        self.logger = logging.getLogger('test.api')
        self.logger.debug("log")

    def _equals(self, i_d1, i_d2):
        """
        checks if d1 is equal of d2
        """
        if type(i_d1) is not dict:
            d1 = i_d1.to_dict()
        else:
            d1 = i_d1
        if type(i_d2) is not dict:
            d2 = i_d2.to_dict()
        else:
            d2 = i_d2
        for key, value in d1.items():
            if not key in d2:
                self.logger.error("ERROR: Key missing  %s" % key)
                return False
            v2 = d2[key]
            if type(v2) is _DictContent and value:
                if not self._equals(value.to_dict(), v2.to_dict()):
                    return False
            else:
                if not v2 == value:
                    # last update is 99% of the time different, especially after an update
                    if key != 'last_update':
                        self.logger.error(
                            "Value error %s: %s %s " % (key, value, v2))
                        return False
        for key in d2.keys():
            if not key in d1:
                self.logger.error("ERROR: Key missing  %s" % key)
                return False
        return True

    def _has_keys(self, keys, d):
        """
        checks if the dictionary has the keys
        """
        for key in keys:
            if not key in d:
                # self.logger.debug("Key missing  %s: %s" % (key, d))
                self.logger.error("ERROR: Key is missing %s" % key)
                return False
        for key in d.keys():
            if not key in keys:
                self.logger.error(
                    "ERROR: Dictionary has an extra key %s" % key)
                return False

        return True


class UserChinoTest(BaseChinoTest):
    user = None

    def setUp(self):
        super(UserChinoTest, self).setUp()
        fields = [dict(name='first_name', type='string'),
                  dict(name='last_name', type='string'),
                  dict(name='email', type='string'),
                  dict(name='age', type='integer', default=None, indexed=True),
                  dict(name='city', type='string', default=None, indexed=True),
                  ]
        self.us = self.chino.user_schemas.create('test', fields)
        self.app_list = []

    def tearDown(self):
        # if user has been created we remove it.
        self.logger.debug("tearing down %s", self.user)
        self.chino.auth.set_auth_admin()
        list = self.chino.users.list(self.us._id)
        for user in list.users:
            self.chino.users.delete(user._id, force=True)
        self.chino.user_schemas.delete(self.us._id, force=True)
        if hasattr(self, 'app'):
            self.chino.applications.delete(self.app._id)

        # delete also every Application which was created forthis test
        for app in self.chino.applications.list().to_dict()['applications']:
            self.chino.applications.delete(app['app_id'], force=True)

    def test_list(self):
        list = self.chino.users.list(self.us._id)
        self.assertIsNotNone(list.paging)
        self.assertIsNotNone(list.users)

    def test_CRUD(self):
        NAME = 'test.user.new'
        EDIT = NAME + '.edited'
        user = self.chino.users.create(self.us._id, username=NAME,
                                       password='12345678',
                                       attributes=dict(first_name='john',
                                                       last_name='doe',
                                                       email='test@chino.io'))
        self.user = user
        self.assertIsNotNone(user)
        self.assertEqual(user.username, NAME)
        list = self.chino.users.list(self.us._id)
        # NOTE: this may fail if key are used by someone else.
        ste_2 = self.chino.users.detail(list.users[0]._id)
        self.assertEqual(ste_2.username, NAME)

        user.username = EDIT
        # remove extra params
        data = user.to_dict()
        del data['insert_date']
        del data['last_update']
        del data['groups']
        del data['schema_id']
        ste_2 = self.chino.users.update(**data)
        self.logger.debug(ste_2)
        self.assertEqual(user._id, ste_2._id)
        self.assertEqual(ste_2.username, EDIT)

        ste_2 = self.chino.users.detail(user._id)
        self.assertEqual(ste_2.username, EDIT)

        # partial update
        ste_3 = self.chino.users.partial_update(user._id, **dict(
            attributes=dict(first_name=EDIT)))
        self.assertEqual(user._id, ste_3._id)
        self.assertEqual(ste_3.attributes.first_name, EDIT)
        ste_4 = self.chino.users.detail(user._id)
        self.assertEqual(ste_4.attributes.first_name, EDIT)
        # current not working for main user
        self.assertRaises(CallError, self.chino.users.current)

        # this invalidates the user
        self.chino.users.delete(user._id)

    def test_search(self):
        # add two users
        data = {"first_name": "John", "last_name": "Doe",
                "email": "test+1@chino.io", "age": 42, "city": "Berlin"}
        user_1 = self.chino.users.create(self.us._id, username="user_1",
                                         password="12345678", attributes=data)
        data = {"first_name": "John", "last_name": "Smith",
                "email": "test+2@chino.io", "age": 21, "city": "Rome"}
        user_2 = self.chino.users.create(self.us._id, username="user_2",
                                         password="12345678", attributes=data)

        # wait for users to be indexed
        time.sleep(1)

        # search
        # - all the users with city == 'Berlin'
        query = {"field": "city", "type": "eq", "value": "Berlin"}
        r = self.chino.users.search(self.us._id, query=query)
        self.assertEquals(1, len(r.users))
        self.assertEquals(user_1._id, r.users[0]._id)

        # - all the users with age <= 30
        query = {"field": "age", "type": "lte", "value": 30}
        r = self.chino.users.search(self.us._id, query=query)
        self.assertEquals(1, len(r.users))
        self.assertEquals(user_2._id, r.users[0]._id)

        # - all the users with age < 30 and city == 'Berlin'
        query = {"and": [
            {"field": "age", "type": "lte", "value": 30},
            {"field": "city", "type": "eq", "value": "Berlin"}
        ]}
        r = self.chino.users.search(self.us._id, query=query)
        self.assertEquals(0, len(r.users))

        # - all the users with age < 30 or city == 'Berlin' order by age desc
        query = {"or": [
            {"field": "age", "type": "lte", "value": 30},
            {"field": "city", "type": "eq", "value": "Berlin"}
        ]}
        sort = [{"field": "age", "order": "desc"}]
        r = self.chino.users.search(self.us._id, query=query, sort=sort)
        self.assertEquals(2, len(r.users))
        self.assertEquals(user_1._id, r.users[0]._id)
        self.assertEquals(user_2._id, r.users[1]._id)

        # - same search with inverted order
        sort = [{"field": "age", "order": "asc"}]
        r = self.chino.users.search(self.us._id, query=query, sort=sort)
        self.assertEquals(2, len(r.users))
        self.assertEquals(user_2._id, r.users[0]._id)
        self.assertEquals(user_1._id, r.users[1]._id)

    # @unittest.skip("method disabled locally")
    def test_auth(self):
        # login
        NAME = 'test.user.new'
        EDIT = NAME + '.edited'
        self.app = self.chino.applications.create("test",
                                                  grant_type='password')
        self.chino_user = ChinoAPIClient(customer_id=cfg.customer_id,
                                         customer_key=cfg.customer_key,
                                         url=cfg.url,
                                         client_id=self.app.app_id,
                                         client_secret=self.app.app_secret,
                                         force_https=False)
        user = self.chino.users.create(self.us._id, username=EDIT,
                                       password='12345678',
                                       attributes=dict(first_name='john',
                                                       last_name='doe',
                                                       email='test@chino.io'))
        self.chino_user.users.login(EDIT, '12345678')
        ste_2 = self.chino_user.users.current()
        self.assertEqual(ste_2.username, EDIT)

        self.chino_user.users.refresh()
        # it should be impossible to create the user after login with self.chino_user (no admin access)
        self.assertRaises(CallError, self.chino_user.users.create, self.us._id,
                          username='error', password='12345678',
                          attributes=dict(first_name='john', last_name='doe',
                                          email='test@chino.io'))

        self.chino_user.users.logout()
        self.assertRaises(Exception, self.chino_user.users.login, EDIT, '')

        self.assertRaises(CallError, self.chino.users.current)

    def test_auth_public(self):
        # login
        NAME = 'test.user.new'
        EDIT = NAME + '.edited'
        self.app = self.chino.applications.create("test",
                                                  grant_type='password',
                                                  client_type='public')
        # Init 'public' client
        self.chino_user = ChinoAPIClient(customer_id=cfg.customer_id,
                                         url=cfg.url,
                                         client_id=self.app.app_id,
                                         client_secret=None,
                                         force_https=False
                                         )
        self.assertIsNone(self.chino_user.auth.client_secret)

        user = self.chino.users.create(self.us._id, username=EDIT,
                                       password='12345678',
                                       attributes=dict(first_name='john',
                                                       last_name='doe',
                                                       email='test@chino.io'))

        self.chino_user.users.login(EDIT, '12345678')
        ste_2 = self.chino_user.users.current()
        self.assertEqual(ste_2.username, EDIT)

        self.chino_user.users.refresh()
        # it should be impossible to create the user after login with self.chino_user (no admin access)
        self.assertRaises(CallError, self.chino_user.users.create, self.us._id,
                          username='error', password='12345678',
                          attributes=dict(first_name='john', last_name='doe',
                                          email='test@chino.io'))

        self.chino_user.users.logout()
        self.assertRaises(Exception, self.chino_user.users.login, EDIT, '')

        self.assertRaises(CallError, self.chino.users.current)


# @unittest.skip("Class disabled locally")
class ApplicationsChinoTest(BaseChinoTest):
    user = None

    def setUp(self):
        super(ApplicationsChinoTest, self).setUp()

    def tearDown(self):
        # if user has been created we remove it.
        self.logger.debug("tearing down %s", self.user)

    def test_CRUD(self):
        app = self.chino.applications.create(name='tesssst_confidential',
                                             grant_type='password')
        app_public = self.chino.applications.create(name='test_public',
                                                    grant_type='password',
                                                    client_type='public')

        app_public1 = self.chino.applications.detail(app_public._id)
        self.assertEqual(app_public._id, app_public1._id)
        self.assertEqual(app_public.app_name, app_public1.app_name)

        newname = 'test_confidential'
        self.chino.applications.update(app._id, name=newname)
        app1 = self.chino.applications.detail(app._id)
        self.assertEqual(app1.app_name, newname)

        apps = self.chino.applications.list()
        self.chino.applications.delete(app_public1._id, force=True)
        self.chino.applications.delete(app1._id, force=True)


# @unittest.skip("Class disabled locally")
class GroupChinoTest(BaseChinoTest):
    def tearDown(self):
        list = self.chino.groups.list()
        for group in list.groups:
            self.chino.groups.delete(group._id, force=True)

        list = self.chino.user_schemas.list()
        for schema in list.user_schemas:
            self.chino.user_schemas.delete(schema._id, force=True)

    def test_list(self):
        list = self.chino.groups.list()
        self.assertIsNotNone(list.paging)
        self.assertIsNotNone(list.groups)

    def test_list_filter(self):
        # create three groups with different names
        now = int(time.time())
        names = ['%s abc1' % now, '%s abc2' % now, 'other %s' % now]
        for name in names:
            self.chino.groups.create(name)

        # search repos with '<now>' in the description: should be 3
        list_all = self.chino.groups.list(name=now)
        self.assertEquals(len(list_all.groups), 3)

        # search repos with '<now> abc' in the description
        list_abc = self.chino.groups.list(name='%s abc' % now)
        self.assertEquals(len(list_abc.groups), 2)

        # search repos with '<now> other' in the description
        list_other = self.chino.groups.list(name='other %s' % now)
        self.assertEquals(len(list_other.groups), 1)

    def test_CRUD(self):
        group_created = self.chino.groups.create('testing', attributes=dict(
            hospital='test'))
        self.assertTrue(self._equals(group_created.attributes.to_dict(),
                                     dict(hospital='test')))
        self.assertEqual(group_created.group_name, 'testing')
        # NOTE: this may fail
        list = self.chino.groups.list()
        group = list.groups[0]
        details = self.chino.groups.detail(group._id)
        # print group.to_dict()
        # print details.to_dict()
        self.assertTrue(self._equals(details, group), "\n %s \n %s \n" % (
        details.to_json(), group.to_json()))
        group.group_name = 'updatedtesting'
        # remove extra params
        data = group.to_dict()
        del data['insert_date']
        del data['last_update']
        self.chino.groups.update(**data)
        details = self.chino.groups.detail(group._id)
        self.assertTrue(self._equals(details, group), "\n %s \n %s \n" % (
        details.to_json(), group.to_json()))
        self.chino.groups.delete(details._id)

    def test_group_user(self):
        group_created = self.chino.groups.create('testing', attributes=dict(
            hospital='test'))
        fields = [dict(name='first_name', type='string'),
                  dict(name='last_name', type='string'),
                  dict(name='email', type='string')]
        us = self.chino.user_schemas.create('test', fields)

        user = self.chino.users.create(us._id, username="ste",
                                       password='12345678',
                                       attributes=dict(first_name='john',
                                                       last_name='doe',
                                                       email='test@chino.io'))
        self.user = user
        # if nothing is raised, then fine
        add = self.chino.groups.add_user(group_created._id, user._id)
        users = self.chino.groups.list_users(group_created._id)
        self.assertEqual(users.paging.count, 1)
        users = self.chino.groups.list_users(group_created._id, offset=2)
        self.assertEqual(users.paging.count, 0)
        rem = self.chino.groups.del_user(group_created._id, user._id)
        # delete the user at the end
        self.chino.groups.delete(group_created._id, force=True)
        self.chino.users.delete(user._id, force=True)


class RepositoryChinoTest(BaseChinoTest):
    def setUp(self):
        super(RepositoryChinoTest, self).setUp()
        list = self.chino.repositories.list()
        for repo in list.repositories:
            self.chino.repositories.delete(repo._id, force=True,
                                           all_content=True)

    def tearDown(self):
        list = self.chino.repositories.list()
        for repo in list.repositories:
            self.chino.repositories.delete(repo._id, force=True,
                                           all_content=True)

    def test_list(self):
        list = self.chino.repositories.list()
        self.assertIsNotNone(list.paging)
        self.assertIsNotNone(list.repositories)

    def test_list_filter(self):
        # create three repositories with different descriptions
        now = int(time.time())
        descriptions = ['%s abc1' % now, '%s abc2' % now, 'other %s' % now]
        repos = [self.chino.repositories.create(d) for d in descriptions]

        # search repos with '<now>' in the description: should be 3
        list_all = self.chino.repositories.list(descr=now)
        self.assertEquals(len(list_all.repositories), 3)

        # search repos with '<now> abc' in the description
        list_abc = self.chino.repositories.list(descr='%s abc' % now)
        self.assertEquals(len(list_abc.repositories), 2)

        # search repos with '<now> other' in the description
        list_other = self.chino.repositories.list(descr='other %s' % now)
        self.assertEquals(len(list_other.repositories), 1)

    def test_crud(self):
        created = self.chino.repositories.create('test')
        self.assertEqual(created.description, 'test')
        first = self.chino.repositories.list().repositories[0]
        self.assertTrue(
            self._equals(created.to_dict(), first.to_dict()),
            "\n %s \n %s \n" % (created.to_json(), first.to_json()))
        first.description = 'edited'

        resp = self.chino.repositories.update(first._id,
                                              description=first.description)
        detail = self.chino.repositories.detail(first._id)
        self.assertTrue(
            self._equals(resp.to_dict(), detail.to_dict()),
            "\n %s \n %s \n" % (resp.to_json(), detail.to_json()))

        self.chino.repositories.delete(first._id)


class SchemaChinoTest(BaseChinoTest):
    def setUp(self):
        super(SchemaChinoTest, self).setUp()
        self.repo = self.chino.repositories.create('test')._id

    def tearDown(self):
        list = self.chino.schemas.list(self.repo)
        for schema in list.schemas:
            self.chino.schemas.delete(schema._id, force=True, all_content=True)
        self.chino.repositories.delete(self.repo, True)

    def test_list(self):
        list = self.chino.schemas.list(self.repo)
        self.assertIsNotNone(list.paging)
        self.assertIsNotNone(list.schemas)

    def test_list_filter(self):
        # create three schemas with different descriptions
        now = int(time.time())
        descriptions = ['%s abc 1' % now, '%s abc 2' % now, 'other %s' % now]
        struct = [{"name": "test", "type": "string"}]
        for description in descriptions:
            self.chino.schemas.create(self.repo, description, fields=struct)

        # now search for them using filter
        list_all = self.chino.schemas.list(self.repo, descr=now)
        self.assertEquals(len(list_all.schemas), 3)
        list_abc = self.chino.schemas.list(self.repo, descr='%s abc' % now)
        self.assertEquals(len(list_abc.schemas), 2)
        list_other = self.chino.schemas.list(self.repo, descr='other %s' % now)
        self.assertEquals(len(list_other.schemas), 1)

    def test_crud(self):
        fields = [dict(name='fieldInt', type='integer'),
                  dict(name='fieldString', type='string'),
                  dict(name='fieldBool', type='boolean'),
                  dict(name='fieldDate', type='date'),
                  dict(name='fieldDateTime', type='datetime')]
        created = self.chino.schemas.create(self.repo, 'test', fields)
        list = self.chino.schemas.list(self.repo)
        detail = self.chino.schemas.detail(list.schemas[0]._id)
        detail2 = self.chino.schemas.detail(created._id)
        self.assertTrue(
            self._equals(detail.to_dict(), detail2.to_dict()),
            "\n %s \n %s \n" % (detail.to_json(), detail2.to_json()))

        detail.structure.fields.append(_Field('string', 'new one'))
        data = detail.to_dict()
        del data['repository_id']
        del data['insert_date']
        del data['last_update']
        self.chino.schemas.update(**data)
        detail2 = self.chino.schemas.detail(detail._id)
        self.assertTrue(
            self._equals(detail.to_dict(), detail2.to_dict()),
            "\n %s \n %s \n" % (detail.to_json(), detail2.to_json()))
        self.chino.schemas.delete(detail._id)

    def test_convert_datatype(self):
        # create a schema
        fields = [{'name': 'numeric', 'type': 'float'}]
        created = self.chino.schemas.create(self.repo, 'test_ct', fields)

        # convert the field to int
        new_fields = {'numeric': 'integer'}
        defaults = {'numeric': 42}
        res = self.chino.schemas.convert_datatypes_submit(created.schema_id,
                                                          fields=new_fields,
                                                          defaults=defaults)
        self.assertIn('message', res)

        # now check the status
        status = self.chino.schemas.convert_datatypes_status(created.schema_id)
        keys = ['start', 'end', 'status']
        for key in keys:
            self.assertIn(key, status)

    def test_dump(self):
        # create a schema
        fields = [{'name': 'numeric', 'type': 'integer', 'indexed': True}]
        created = self.chino.schemas.create(self.repo, 'test_dump', fields)

        # add 10 documents
        for i in range(10):
            record = {'numeric': i}
            self.chino.documents.create(created.schema_id, record,
                                        consistent=True)

        # submit the dump job
        query = {"field": "numeric", "type": "lt", "value": 5}
        sort = [{"field": "numeric", "order": "asc"}]
        res = self.chino.schemas.dump_submit(created.schema_id, query=query,
                                             sort=sort)
        self.assertIn('dump_id', res)
        dump_id = res['dump_id']

        timeout = 10
        while timeout:
            res = self.chino.schemas.dump_status(dump_id)
            if res['status'] == 'COMPLETED':
                break
            time.sleep(1)
            timeout -= 1

        self.assertEqual(res['status'], 'COMPLETED')

        # now download the file
        dump = self.chino.schemas.dump_download(dump_id)

        # check that the file is not corrupted
        self.assertTrue(dump.hash_matches())

        # should keep 5 records
        self.assertEqual(5, dump.records_count)
        self.assertEqual(5, len(dump.json))

        # check the records' content
        for i, record in enumerate(dump.json):
            self.assertEqual(i, record['numeric'])

        # now test the csv format
        res = self.chino.schemas.dump_submit(created.schema_id, query=query,
                                             sort=sort, format='csv')
        self.assertIn('dump_id', res)
        dump_id = res['dump_id']

        timeout = 10
        while timeout:
            res = self.chino.schemas.dump_status(dump_id)
            if res['status'] == 'COMPLETED':
                break
            time.sleep(1)
            timeout -= 1

        self.assertEqual(res['status'], 'COMPLETED')

        # now download the file
        dump = self.chino.schemas.dump_download(dump_id)

        # check that the file is not corrupted
        self.assertTrue(dump.hash_matches())

        # should keep 5 records + 1 line of headers
        self.assertEqual(5, dump.records_count)
        self.assertEqual(6, len(dump.csv))

        # check the records' content, ordered
        for i, record in enumerate(dump.csv):
            if i == 0:
                self.assertEqual('_id', record[0])
                self.assertEqual('numeric', record[1])
            else:
                self.assertEqual(str(i - 1), record[1])


class UserSchemaChinoTest(BaseChinoTest):
    def setUp(self):
        super(UserSchemaChinoTest, self).setUp()

    def tearDown(self):
        list = self.chino.user_schemas.list()
        for repo in list.user_schemas:
            self.chino.user_schemas.delete(repo._id, force=True)

    def test_list(self):
        list = self.chino.user_schemas.list()
        self.assertIsNotNone(list.paging)
        self.assertIsNotNone(list.user_schemas)

    def test_list_filter(self):
        # create three user schemas with different descriptions
        now = int(time.time())
        descriptions = ['%s abc 1' % now, '%s abc 2' % now, 'other %s' % now]
        for description in descriptions:
            self.chino.user_schemas.create(description, [])

        # now search for them using filter
        list_all = self.chino.user_schemas.list(descr=now)
        self.assertEquals(len(list_all.user_schemas), 3)
        list_abc = self.chino.user_schemas.list(descr='%s abc' % now)
        self.assertEquals(len(list_abc.user_schemas), 2)
        list_other = self.chino.user_schemas.list(descr='other %s' % now)
        self.assertEquals(len(list_other.user_schemas), 1)

    def test_crud(self):
        fields = [dict(name='fieldInt', type='integer'),
                  dict(name='fieldString', type='string'),
                  dict(name='fieldBool', type='boolean'),
                  dict(name='fieldDate', type='date'),
                  dict(name='fieldDateTime', type='datetime')]
        created = self.chino.user_schemas.create('test', fields)
        list = self.chino.user_schemas.list()
        self.assertGreater(list.paging.count, 0)
        id = 0
        for schema in list.user_schemas:
            if schema._id == created._id:
                id = schema._id
        detail = self.chino.user_schemas.detail(id)
        detail2 = self.chino.user_schemas.detail(created._id)
        self.assertTrue(
            self._equals(detail.to_dict(), detail2.to_dict()),
            "\n %s \n %s \n" % (detail.to_json(), detail2.to_json()))

        detail.structure.fields.append(_Field('string', 'new one'))
        data = detail.to_dict()
        del data['insert_date']
        del data['last_update']
        del data['user_schema_id']
        del data['groups']
        self.chino.user_schemas.update(detail.user_schema_id, **data)
        detail2 = self.chino.user_schemas.detail(detail._id)
        self.assertTrue(
            self._equals(detail.to_dict(), detail2.to_dict()),
            "\n %s \n %s \n" % (detail.to_json(), detail2.to_json()))
        self.chino.user_schemas.delete(detail._id, force=True)


class CollectionChinoTest(BaseChinoTest):
    def setUp(self):
        super(CollectionChinoTest, self).setUp()

        self.repo = self.chino.repositories.create('test')._id
        fields = [dict(name='fieldInt', type='integer'),
                  dict(name='fieldString', type='string'),
                  dict(name='fieldBool', type='boolean'),
                  dict(name='fieldDate', type='date'),
                  dict(name='fieldDateTime', type='datetime')]
        self.schema = self.chino.schemas.create(self.repo, 'test', fields)._id
        fields = [dict(name='fieldString', type='text')]

    def tearDown(self):
        list = self.chino.collections.list()
        for coll in list.collections:
            self.chino.collections.delete(coll._id, force=True)

        list = self.chino.documents.list(self.schema)
        for doc in list.documents:
            self.chino.documents.delete(doc._id, force=True)
        self.chino.schemas.delete(self.schema, True)
        self.chino.repositories.delete(self.repo, force=True, all_content=True)

    def test_list(self):
        list = self.chino.user_schemas.list()
        self.assertIsNotNone(list.paging)
        self.assertIsNotNone(list.user_schemas)

    def test_list_filter(self):
        # create a collection and two documents. Add one doc to the collection
        # then list the collections filtering by document

        # create the collection
        coll = self.chino.collections.create("test_")

        # create the docs
        content = dict(fieldInt=123, fieldString='test', fieldBool=False,
                       fieldDate='2015-02-19',
                       fieldDateTime='2015-02-19T16:39:47')

        doc1 = self.chino.documents.create(self.schema, content=content)
        doc2 = self.chino.documents.create(self.schema, content=content)

        # add the doc1 to the collection
        self.chino.collections.add_document(coll._id, doc1._id)

        # list the collections filtering by doc1: should be 1
        list_1 = self.chino.collections.list(document_id=doc1._id)
        self.assertEquals(len(list_1.collections), 1)

        # list the collections filtering by doc2: should be 0
        list_2 = self.chino.collections.list(document_id=doc2._id)
        self.assertEquals(len(list_2.collections), 0)

    def test_crud(self):
        created = self.chino.collections.create("test_")
        list = self.chino.collections.list()
        detail = self.chino.collections.detail(list.collections[0]._id)
        detail2 = self.chino.collections.detail(created._id)
        self.assertTrue(
            self._equals(detail.to_dict(), detail2.to_dict()),
            "\n %s \n %s \n" % (detail.to_json(), detail2.to_json()))
        self.chino.collections.update(detail._id, name='test2')
        detail2 = self.chino.collections.detail(detail._id)
        self.assertTrue(detail2.name == 'test2')
        self.chino.collections.delete(detail._id)

    def test_search(self):
        ids = []
        for i in range(10):
            created = self.chino.collections.create("test" + str(i))
            ids.append(created._id)

        res = self.chino.collections.search('test', contains=True)
        self.assertEqual(res.paging.total_count, 10)
        res = self.chino.collections.search('test2', contains=False)
        for collection in res.collections:
            self.assertTrue(collection.name.startswith('test'))
        self.assertEqual(res.paging.total_count, 1)
        self.assertEqual(res.collections[0].name, 'test2')
        for id in ids:
            self.chino.collections.delete(id, True)

    def test_docs(self):
        repo = self.chino.repositories.create('test')._id
        fields = [dict(name='fieldInt', type='integer'),
                  dict(name='fieldString', type='string'),
                  dict(name='fieldBool', type='boolean'),
                  dict(name='fieldDate', type='date'),
                  dict(name='fieldDateTime', type='datetime')]
        schema = self.chino.schemas.create(repo, 'test', fields)._id
        content = dict(fieldInt=123, fieldString='test', fieldBool=False,
                       fieldDate='2015-02-19',
                       fieldDateTime='2015-02-19T16:39:47')
        document = self.chino.documents.create(schema, content=content)
        collection = self.chino.collections.create("test")
        l = self.chino.collections.list_documents(collection._id)
        self.assertEqual(l.paging.count, 0)
        self.chino.collections.add_document(collection._id, document._id)
        l = self.chino.collections.list_documents(collection._id)
        self.assertEqual(l.paging.count, 1)
        self._equals(l.documents[0].to_dict(), document.to_dict())
        self.chino.collections.rm_document(collection._id, document._id)
        l = self.chino.collections.list_documents(collection._id)
        self.assertEqual(l.paging.count, 0)

        # delete
        self.chino.documents.delete(document._id, force=True)
        self.chino.schemas.delete(schema, force=True)
        self.chino.repositories.delete(repo, force=True)
        self.chino.collections.delete(collection._id, force=True)


@unittest.skip("Test to be updated")
class PermissionChinoTest2(BaseChinoTest):
    def setUp(self):
        super(PermissionChinoTest2, self).setUp()
        # list = self.chino.users.list()
        # for u in list.users:
        # self.chino.users.delete(u._id,force=True)

    def test_premissions(self):
        # create user via userschema
        repo = self.chino.repositories.create('test')._id
        fields = [dict(name='fieldInt', type='integer'),
                  dict(name='fieldString', type='string'),
                  dict(name='fieldBool', type='boolean'),
                  dict(name='fieldDate', type='date'),
                  dict(name='fieldDateTime', type='datetime')]
        schema = self.chino.schemas.create(repo, 'test', fields)._id
        # schema = "c0d0d956-8cd1-405b-90a9-62d3b3f70e84"
        content = dict(fieldInt=123, fieldString='test', fieldBool=False,
                       fieldDate='2015-02-19',
                       fieldDateTime='2015-02-19T16:39:47')
        document = self.chino.documents.create(schema, content=content)
        fields = [dict(name='first_name', type='string'),
                  dict(name='last_name', type='string'),
                  dict(name='email', type='string')]
        us = self.chino.user_schemas.create('test', fields)
        username = 'test_%s' % round(time.time())
        user = self.chino.users.create(us._id, username=username,
                                       password='12345678',
                                       attributes=dict(first_name='john',
                                                       last_name='doe',
                                                       email='test@chino.io'))
        self.chino.permissions.resources('grant', 'repositories', 'users',
                                         user._id, manage=['R'])
        self.chino.users.login(username, '12345678')
        permissions = self.chino.permissions.read_perms()
        self.assertTrue(permissions[0].permission.manage == ['R'])
        self.chino.users.logout()
        self.chino.auth.set_auth_admin()
        self.chino.permissions.resource('grant', 'documents', document._id,
                                        'users', user._id, manage=['R', 'U'])
        self.chino.users.login(username, '12345678')
        permissions = self.chino.permissions.read_perms_document(document._id)
        self.assertTrue(permissions[0].permission.manage == ['R', 'U'])
        self.chino.users.logout()
        self.chino.auth.set_auth_admin()
        self.chino.permissions.resource_children('grant', 'schemas', schema,
                                                 'documents', 'users',
                                                 user._id,
                                                 manage=['R', 'U', 'L'],
                                                 authorize=['A'])
        self.chino.users.login(username, '12345678')
        permissions = self.chino.permissions.read_perms()
        # 0 is the one above
        self.assertTrue(permissions[1].permission.manage == ['R', 'U', 'L'])
        self.assertTrue(permissions[1].permission.authorize == ['A'])

        permissions = self.chino.permissions.read_perms_user(user._id)
        self.assertTrue(permissions[0].permission.manage == ['R'])
        self.assertTrue(permissions[1].permission.manage == ['R', 'U', 'L'])
        self.assertTrue(permissions[1].permission.authorize == ['A'])
        self.chino.users.logout()
        self.chino.auth.set_auth_admin()
        self.chino.documents.delete(document._id, force=True)
        self.chino.schemas.delete(schema, force=True)
        self.chino.repositories.delete(repo, force=True)
        self.chino.users.delete(user._id, force=True)
        self.chino.user_schemas.delete(us._id, force=True)


class DocumentChinoTest(BaseChinoTest):
    def setUp(self):
        super(DocumentChinoTest, self).setUp()
        self.repo = self.chino.repositories.create('test')._id
        fields = [dict(name='fieldInt', type='integer', indexed=True),
                  dict(name='fieldString', type='string'),
                  dict(name='fieldBool', type='boolean', indexed=True),
                  dict(name='fieldDate', type='date'),
                  dict(name='fieldDateTime', type='datetime')]
        self.schema = self.chino.schemas.create(self.repo, 'test', fields)._id
        fields = [dict(name='fieldString', type='text')]
        self.schema_1 = self.chino.schemas.create(self.repo, 'test',
                                                  fields)._id

    def tearDown(self):
        list = self.chino.documents.list(self.schema)
        for doc in list.documents:
            self.chino.documents.delete(doc._id, force=True)
        self.chino.schemas.delete(self.schema, force=True, all_content=True)
        self.chino.repositories.delete(self.repo, force=True, all_content=True)

    def test_list(self):
        content = dict(fieldInt=123, fieldString='test', fieldBool=False,
                       fieldDate='2015-02-19',
                       fieldDateTime='2015-02-19T16:39:47')
        document = self.chino.documents.create(self.schema, content=content)
        list = self.chino.documents.list(self.schema)
        self.assertIsNotNone(list.paging)
        self.assertIsNotNone(list.documents)
        self.assertIsNone(list.documents[0].content)
        list = self.chino.documents.list(self.schema, True)
        self.assertIsNotNone(list.documents[0].content)
        self.chino.documents.delete(document._id, True)

    def test_list_metadata_filters(self):
        # add a document
        content = dict(fieldInt=123, fieldString='test', fieldBool=False,
                       fieldDate='2015-02-19',
                       fieldDateTime='2015-02-19T16:39:47')
        document = self.chino.documents.create(self.schema, content=content)

        # we should get 1 doc
        doc_list = self.chino.documents.list(self.schema)
        self.assertEqual(1, len(doc_list.documents))

        # input('waiting (strike any key to continue)')
        time.sleep(1)
        now = datetime.now(tz=pytz.timezone('Europe/Berlin'))

        # document - transform strings to datetime objects
        doc_insert_date = datetime.strptime(
            document.insert_date,
            "%Y-%m-%dT%H:%M:%S.%fZ"
        )
        doc_last_update = datetime.strptime(
            document.last_update,
            "%Y-%m-%dT%H:%M:%S.%fZ"
        )

        filters = [
            ({"is_active": True}, 1),
            ({"is_active": False}, 0),
            ({"is_active": True,
              "last_update__gt": doc_last_update + timedelta(seconds=1)}, 0),
            ({"is_active": True,
              "insert_date__gt": doc_insert_date - timedelta(seconds=1)}, 1),
            ({"last_update__lt": doc_last_update + timedelta(hours=1)}, 1),
            ({"insert_date__lt": doc_insert_date + timedelta(hours=1)}, 1),
            ({"is_active": True, "insert_date__lt": now,
              "last_update__lt": now}, 1),
            ({"is_active": True, "insert_date__gt": now,
              "last_update__gt": now}, 0),
        ]
        # we build a list of possible combination of filters - expected values
        for params, expected_count in filters:
            result = self.chino.documents.list(self.schema, **params)
            self.assertEqual(expected_count, len(result.documents), params)
            if expected_count:
                self.assertEqual(document._id, result.documents[0]._id)

    def test_list_metadata_ordering(self):
        # add 3 documents
        content = dict(fieldInt=123, fieldString='test', fieldBool=False,
                       fieldDate='2015-02-19',
                       fieldDateTime='2015-02-19T16:39:47')

        docs = []
        for i in range(3):
            doc = self.chino.documents.create(self.schema, content=content)
            docs.append(doc)
            time.sleep(1)

        # now we edit the first one, in order to change its last_update_value
        first = docs[0]
        content['fieldString'] = 'changed'
        self.chino.documents.update(first._id, content=content)

        # now we list em by insert_date and we check ordering
        params = {'ordering': 'insert_date'}
        res = self.chino.documents.list(self.schema, **params)

        for i, item in enumerate(res.documents):
            self.assertEqual(docs[i]._id, item._id)

        # invert order
        params = {'ordering': '-insert_date'}
        res = self.chino.documents.list(self.schema, **params)
        for i, item in enumerate(res.documents):
            self.assertEqual(docs[-i-1]._id, item._id)

        # now we check the order by last_update
        params = {'ordering': 'last_update'}
        res = self.chino.documents.list(self.schema, **params)
        docs_last_update = [docs[1], docs[2], docs[0]]
        for i, item in enumerate(res.documents):
            self.assertEqual(docs_last_update[i]._id, item._id)

        # invert order
        params = {'ordering': '-last_update'}
        res = self.chino.documents.list(self.schema, **params)
        for i, item in enumerate(res.documents):
            self.assertEqual(docs_last_update[-i-1]._id, item._id)

        # combine ordering
        params = {'ordering': 'insert_date,-last_update'}
        res = self.chino.documents.list(self.schema, **params)
        docs_combo = [docs[0], docs[1], docs[2]]
        for i, item in enumerate(res.documents):
            self.assertEqual(docs_combo[i]._id, item._id)

    # @unittest.skip("not working, timeout`")
    def test_too_big(self):
        k_bit = 'YXNkc2FkamtzZGprYWhqa3NkaGFqa3NoZGpzYWhkamtzYWhkamtoc2Fqa2xoamRrc2ZsaGpka2xzaGZhamtkbHNoamFrZmxoZGp' \
                'za2xhaGZqZGtsc2hqZmtsYWhqZmtkbGhzYWpma2xkaHNqa2xhaGZqZGtsc2hhamZrbGFoamtkc2xoamZrZGxzaGpma2xkc2hham' \
                'tmbGRoc2pha2xmaGpka2xzaGpma2FzZGRqc2FpbyBkanNpYW9qZGlzYWpkaW9zYWpkaW9zYWppZG9qc2Fpb2RqaXNhb2pkaXNvY' \
                'WpkaW9zamFpZG9qc2Fpb2RqaXcxMDkzMDgyMTkwOGRtMXdpMDltZGk5MDFtaTlkMG13MWk5MGRtaXc5MTBtaWQ5dzBtaGRzamFr' \
                'aGRqa3NoYWRqbGFoZnNqa2xkaHNhamZrbGRoc2pha2xoZmpkc2tsYWhmamRrc2xoZmllYXVoZmlvZWgyYXVpb2ZoZXVpb3BxamV' \
                'pb3dlangsYXNhZHNhc2RzYWRqa3NkamthaGprc2RoYWprc2hkanNhaGRqa3NhaGRqa2hzYWprbGhqZGtzZmxoamRrbHNoZmFqa2' \
                'Rsc2hqYWtmbGhkanNrbGFoZmpka2xzaGpma2xhaGpma2RsaHNhamZrbGRoc2prbGFoZmpka2xzaGFqZmtsYWhqa2RzbGhqZmtkb' \
                'HNoamZrbGRzaGFqa2ZsZGhzamFrbGZoamRrbHNoamZrYXNkZGpzYWlvIGRqc2lhb2pkaXNhamRpb3NhamRpb3Nhamlkb2pzYWlv' \
                'ZGppc2FvamRpc29hamRpb3NqYWlkb2pzYWlvZGppdzEwOTMwODIxOTA4ZG0xd2kwOW1kaTkwMW1pOWQwbXcxaTkwZG1pdzkxMG1' \
                'pZDl3MG1oZHNqYWtoZGprc2hhZGpsYWhmc2prbGRoc2FqZmtsZGhzamFrbGhmamRza2xhaGZqZGtzbGhmaWVhdWhmaW9laDJhdW' \
                'lvZmhldWlvcHFqZWlvd2VqeCxhc2Fkc2FzZHNhZGprc2Rqa2FoamtzZGhhamtzaGRqc2FoZGprc2FoZGpraHNhamtsaGpka3Nmb' \
                'GhqZGtsc2hmYWprZGxzaGpha2ZsaGRqc2tsYWhmamRrbHNoamZrbGFoamZrZGxoc2FqZmtsZGhzamtsYWhmamRrbHNoYWpma2xh' \
                'aGprZHNsaGpma2Rsc2hqZmtsZHNoYWprZmxkaHNqYWtsZmhqZGtsc2hqZmthc2RkanNhaW8gZGpzaWFvamRpc2FqZGlvc2FqZGl' \
                'vc2FqaWRvanNhaW9kamlzYW9qZGlzb2FqZGlvc2phaWRvanNhaW9kaml3MTA5MzA4MjE5MDhkbTF3aTA5bWRpOTAxbWk5ZDBtdz' \
                'FpOTBkbWl3OTEwbWlkOXcwbWhkc2pha2hkamtzaGFkamxhaGZzamtsZGhzYWpma2xkaHNqYWtsaGZqZHNrbGFoZmpka3NsaGZpZ' \
                'WF1aGZpb2VoMmF1aW9maGV1aW9wcWplaW93ZWp4LGFzYWRzYXNkc2FkamtzZGprYWhqa3NkaGFqa3NoZGpzYWhkamtzYWhkamto' \
                'c2Fqa2xoamRrc2ZsaGpka2xzaGZhamtkbHNoamFrZmxoZGpza2xhaGZqZGtsc2hqZmtsYWhqZmtkbGhzYWpma2xkaHNqa2xhaGZ' \
                'qZGtsc2hhamZrbGFoamtkc2xoamZrZGxzaGpma2xkc2hhamtmbGRoc2pha2xmaGpka2xzaGpma2FzZGRqc2FpbyBkanNpYW9qZG' \
                'lzYWpkaW9zYWpkaW9zYWppZG9qc2Fpb2RqaXNhb2pkaXNvYWpkaW9zamFpZG9qc2Fpb2RqaXcxMDkzMDgyMTkwOGRtMXdpMDltZ' \
                'Gk5MDFtaTlkMG13MWk5MGRtaXc5MTBtaWQ5dzBtaGRzamFraGRqa3NoYWRqbGFoZnNqa2xkaHNhamZrbGRoc2pha2xoZmpkc2ts' \
                'YWhmamRrc2xoZmllYXVoZmlvZWgyYXVpb2ZoZXVpb3BxamVpb3dlangsYXNhZHNkc2Fh'
        n = 21  # number of MB
        k = 0
        dat = ""
        self.chino.documents.create(self.schema_1,
                                    content=dict(fieldString=dat))
        while k < n * 512:
            dat += "%s" % k_bit
            k += 1
        #
        with self.assertRaises(CallError):  # must fail because is too big
            self.chino.documents.create(self.schema_1,
                                        content=dict(fieldString=dat))

    def test_crud(self):
        content = dict(fieldInt=123, fieldString='test', fieldBool=False,
                       fieldDate='2015-02-19',
                       fieldDateTime='2015-02-19T16:39:47')
        document = self.chino.documents.create(self.schema, content=content)
        # get the last
        document_det = self.chino.documents.detail(document._id)
        self.assertEqual(document.last_update, document_det.last_update)
        content = dict(fieldInt=123,
                       fieldString='test', fieldBool=False,
                       fieldDate='2015-02-19',
                       fieldDateTime='2015-02-19T16:39:47')
        self.chino.documents.update(document._id, content=content)
        document = self.chino.documents.detail(document._id)
        self.assertEqual(123, document.content.fieldInt)
        self.chino.documents.update(document._id,
                                    content=dict(fieldInt=349,
                                                 fieldString='test',
                                                 fieldBool=False,
                                                 fieldDate='2015-02-19',
                                                 fieldDateTime='2015-02-19T16:39:47'))
        document = self.chino.documents.detail(document._id)
        self.assertEqual(349, document.content.fieldInt)

        # partial update
        self.chino.documents.update(document._id, content={'fieldInt': 42},
                                    partial=True)
        document = self.chino.documents.detail(document._id)
        self.assertEqual(42, document.content.fieldInt)
        self.assertEqual('test', document.content.fieldString)

        # now remove the doc
        self.chino.documents.delete(document._id)

        fields = [dict(name='fieldInt', type='integer'),
                  dict(name='fieldString', type='string'),
                  dict(name='fieldBool', type='boolean'),
                  dict(name='fieldDate', type='date'),
                  dict(name='fieldDateTime', type='datetime')]
        created = self.chino.schemas.create(self.repo, 'test', fields)
        list = self.chino.schemas.list(self.repo)
        s_id = 0
        for schema in list.schemas:
            if schema._id == created._id:
                s_id = schema._id
        detail = self.chino.schemas.detail(s_id)
        detail2 = self.chino.schemas.detail(created._id)
        self.assertTrue(
            self._equals(detail.to_dict(), detail2.to_dict()),
            "\n %s \n %s \n" % (detail.to_json(), detail2.to_json()))

        detail.structure.fields.append(_Field('string', 'new one'))
        # delete repository_id
        data = detail.to_dict()
        del data['repository_id']
        del data['insert_date']
        del data['last_update']
        self.chino.schemas.update(**data)
        detail2 = self.chino.schemas.detail(detail._id)
        self.assertTrue(
            self._equals(detail.to_dict(), detail2.to_dict()),
            "\n %s \n %s \n" % (detail.to_json(), detail2.to_json()))
        self.chino.schemas.delete(detail._id, force=True)

    def test_search(self):
        # add two documents
        content = {"fieldInt": 42, "fieldString": "antani", "fieldBool": True,
                   "fieldDate": "2021-03-17",
                   "fieldDateTime": "2021-03-17T17:49:00"}
        doc_1 = self.chino.documents.create(self.schema, content=content)

        content = {"fieldInt": 21, "fieldString": "ant", "fieldBool": False,
                   "fieldDate": "2021-03-17",
                   "fieldDateTime": "2021-03-17T17:49:00"}
        doc_2 = self.chino.documents.create(self.schema, content=content)

        # wait for documents to be indexed
        time.sleep(1)

        # search:
        # - all the docs with bool field True
        query = {"field": "fieldBool", "type": "eq", "value": True}
        r = self.chino.documents.search(self.schema, query=query)
        self.assertEquals(1, len(r.documents))
        self.assertEquals(doc_1._id, r.documents[0]._id)

        # - all the docs with int field < 30
        query = {"field": "fieldInt", "type": "lt", "value": 30}
        r = self.chino.documents.search(self.schema, query=query)
        self.assertEquals(1, len(r.documents))
        self.assertEquals(doc_2._id, r.documents[0]._id)

        # - all the docs with bool field True or int field < 30
        query = {"or": [
            {"field": "fieldBool", "type": "eq", "value": True},
            {"field": "fieldInt", "type": "lt", "value": 30}
        ]}
        r = self.chino.documents.search(self.schema, query=query)
        self.assertEquals(2, len(r.documents))
        found_doc_ids = [doc._id for doc in r.documents]
        self.assertIn(doc_1._id, found_doc_ids)
        self.assertIn(doc_2._id, found_doc_ids)

        # - all the docs with bool field False and int field > 50
        query = {"and": [
            {"field": "fieldBool", "type": "eq", "value": False},
            {"field": "fieldInt", "type": "gt", "value": 50}
        ]}
        r = self.chino.documents.search(self.schema, query=query)
        self.assertEquals(0, len(r.documents))

        # - all the docs with int field > 0 ordered DESC
        query = {"field": "fieldInt", "type": "gt", "value": 0}
        sort = [{"field": "fieldInt", "order": "desc"}]
        r = self.chino.documents.search(self.schema, query=query,
                                        sort=sort)
        self.assertEquals(2, len(r.documents))
        self.assertIn(doc_1._id, r.documents[0]._id)
        self.assertIn(doc_2._id, r.documents[1]._id)

        # - same search with inverted order
        sort = [{"field": "fieldInt", "order": "asc"}]
        r = self.chino.documents.search(self.schema, query=query,
                                        sort=sort)
        self.assertEquals(2, len(r.documents))
        self.assertIn(doc_2._id, r.documents[0]._id)
        self.assertIn(doc_1._id, r.documents[1]._id)


class BlobChinoTest(BaseChinoTest):
    def setUp(self):
        super(BlobChinoTest, self).setUp()
        self.repo = self.chino.repositories.create('test')._id
        fields = [dict(name='blobTest', type='blob'),
                  dict(name='name', type='string')]
        self.schema = self.chino.schemas.create(self.repo, 'test', fields)._id

    def tearDown(self):
        if hasattr(self, 'blob'):
            self.chino.blobs.delete(self.blob.blob_id)
        self.chino.documents.delete(self.document._id, True)
        self.chino.schemas.delete(self.schema, True)
        self.chino.repositories.delete(self.repo, True)

    # @unittest.skip("not working on prod")
    def test_blob(self):
        self.document = self.chino.documents.create(self.schema,
                                                    content=dict(name='test'))
        src = os.path.join(os.path.sep, os.path.dirname(__file__), 'logo.png')
        blob = self.chino.blobs.send(self.document._id, 'blobTest', str(src),
                                     chunk_size=32 * 1024)
        blob_detail = self.chino.blobs.detail(blob.blob_id)
        rw = open("out" + blob_detail.filename, "wb")
        rw.write(blob_detail.content)
        rw.close()
        # rd = open('test/logo.png', "rb")
        md5_detail = hashlib.md5()
        md5_detail.update(blob_detail.content)
        # self.assertEqual(md5_detail.digest(), md5_original.digest())
        self.assertEqual(md5_detail.hexdigest(), blob.md5)
        self.blob = blob
        self.chino.blobs.delete(self.blob.blob_id)
        src = os.path.join(os.path.sep, os.path.dirname(__file__), 'test.sh')
        blob = self.chino.blobs.send(self.document._id, 'blobTest', str(src),
                                     chunk_size=32 * 1024)

        blob_detail = self.chino.blobs.detail(blob.blob_id)
        rw = open("out" + blob_detail.filename, "wb")
        rw.write(blob_detail.content)
        rw.close()
        # rd = open('test/logo.png', "rb")
        md5_detail = hashlib.md5()
        md5_detail.update(blob_detail.content)
        # self.assertEqual(md5_detail.digest(), md5_original.digest())
        self.assertEqual(md5_detail.hexdigest(), blob.md5)
        self.blob = blob

    def test_url(self):
        self.document = self.chino.documents.create(self.schema,
                                                    content=dict(name='test'))
        src = os.path.join(os.path.sep, os.path.dirname(__file__), 'logo.png')
        blob = self.chino.blobs.send(self.document._id, 'blobTest', str(src),
                                     chunk_size=32 * 1024)
        data = self.chino.blobs.generate(blob.blob_id,one_time=True)
        blob_detail = self.chino.blobs.detail_token(blob.blob_id,
                                                    data['token'], 'test.out')
        rw = open("out" + blob_detail.filename, "wb")
        rw.write(blob_detail.content)
        rw.close()
        # rd = open('test/logo.png', "rb")
        md5_detail = hashlib.md5()
        md5_detail.update(blob_detail.content)
        self.assertEqual(md5_detail.hexdigest(), blob.md5)

        with self.assertRaises(CallError):
            blob_detail = self.chino.blobs.detail_token(blob.blob_id,
                                                    data['token'], 'test.out')

        self.chino.blobs.delete(blob.blob_id)


    @unittest.skip("Skipping for size")
    def test_large_blob(self):
        self.document = self.chino.documents.create(self.schema,
                                                    content=dict(name='test'))
        import random
        blob = self.chino.blobs.send(self.document._id, 'blobTest',
                                     'dummy.file',
                                     chunk_size=random.choice(
                                         [1, 10]) * 1024 * 1024)
        blob_detail = self.chino.blobs.detail(blob.blob_id)
        rw = open("out" + blob_detail.filename, "wb")
        rw.write(blob_detail.content)
        rw.close()
        # rd = open('test/logo.png', "rb")
        md5_detail = hashlib.md5()
        md5_detail.update(blob_detail.content)
        sha1_detail = hashlib.sha1()
        sha1_detail.update(blob_detail.content)
        # self.assertEqual(md5_detail.digest(), md5_original.digest())
        self.assertEqual(md5_detail.hexdigest(), blob.md5)
        self.assertEqual(sha1_detail.hexdigest(), blob.sha1)
        # self.blob = blob


# @unittest.skip("not working on prod`")
class SearchDocsChinoTest(BaseChinoTest):
    def setUp(self):
        super(SearchDocsChinoTest, self).setUp()
        fields = [dict(name='fieldInt', type='integer', indexed=True),
                  dict(name='fieldString', type='string', indexed=True),
                  dict(name='fieldBool', type='boolean', indexed=True),
                  dict(name='fieldDate', type='date', indexed=True),
                  dict(name='fieldDateTime', type='datetime', indexed=True)]
        self.repo = self.chino.repositories.create('test')._id
        self.schema = self.chino.schemas.create(self.repo, 'test', fields)._id

    # def tearDown(self):
    #     documents = self.chino.documents.list(self.schema)
    #     for document in documents.documents:
    #         self.chino.documents.delete(document._id, force=True)
    #     self.chino.schemas.delete(self.schema, True)
    #     self.chino.repositories.delete(self.repo, True)

    # DEPRECATED: to be removed in v>=3.0.0
    def test_search_docs(self):
        tot = 9
        print(self.schema)
        for i in range(tot):
            res = self.chino.documents.create(self.schema,
                                              content=dict(fieldInt=123,
                                                           fieldString='test',
                                                           fieldBool=False,
                                                           fieldDate='2015-02-19',
                                                           fieldDateTime='2015-02-19T16:39:47'),
                                              consistent=True)
        last_doc = self.chino.documents.create(self.schema,
                                               content=dict(fieldInt=123,
                                                            fieldString='test',
                                                            fieldBool=False,
                                                            fieldDate='2015-02-19',
                                                            fieldDateTime='2015-02-19T16:39:47'),
                                               consistent=True)

        # self.chino.searches.search(self.schema) # TODO: improve tests
        time.sleep(float(cfg.sleep))  # wait the index max update time
        res = self.chino.searches.documents(self.schema, filters=[
            {"field": "fieldInt", "type": "eq", "value": 123}])
        self.assertEqual(res.paging.total_count, 10, res)
        self.chino.documents.delete(last_doc.document_id, force=True)
        time.sleep(float(cfg.sleep) * 3.0)
        res = self.chino.searches.documents(self.schema, filters=[
            {"field": "fieldInt", "type": "eq", "value": 123}])
        self.assertEqual(res.paging.total_count, 9, res)
        res = self.chino.searches.documents_complex(self.schema,
                                                    query={"field": "fieldInt",
                                                           "type": "eq",
                                                           "value": 123})
        self.assertEqual(res.paging.total_count, 9, res)
        # print res.documents
        ids = []
        for d in res.documents[:3]:
            ids.append(d._id)
        res = self.chino.searches.documents(self.schema, filters=[
            {"field": "_id", "type": "in", "value": ids}])
        self.assertEqual(res.paging.total_count, 3, res)

    # DEPRECATED: to be removed in v>=3.0.0
    def test_search_docs_consistent(self):
        doc = None
        max = 4
        for i in range(max):
            doc = self.chino.documents.create(self.schema,
                                              content=dict(fieldInt=i,
                                                           fieldString='test',
                                                           fieldBool=False,
                                                           fieldDate='2015-02-19',
                                                           fieldDateTime='2015-02-19T16:39:47'),
                                              consistent=True)
        time.sleep(float(cfg.sleep))
        res = self.chino.searches.documents(self.schema,
                                            filters=[{"field": "fieldInt",
                                                      "type": "eq",
                                                      "value": i}])
        self.assertEqual(res.paging.total_count, 1, res)
        res = self.chino.searches.documents(self.schema,
                                            filters=[{"field": "fieldInt",
                                                      "type": "eq",
                                                      "value": max - 1}])
        self.assertEqual(res.paging.total_count, 1, res)
        self.chino.documents.delete(doc._id, consistent=True, force=True)
        res = self.chino.searches.documents(self.schema,
                                            filters=[{"field": "fieldInt",
                                                      "type": "eq",
                                                      "value": max - 1}])
        self.assertEqual(res.paging.total_count, 0, res)

    # DEPRECATED: to be removed in v>=3.0.0
    def test_search_docs_complex(self):
        doc = None
        max = 4
        for i in range(max):
            doc_content = dict(fieldInt=i, fieldString='test', fieldBool=False,
                               fieldDate='2018-12-19',
                               fieldDateTime='2018-12-19T16:39:47')
            doc = self.chino.documents.create(
                self.schema,
                content=doc_content,
                consistent=True
            )
            res = self.chino.searches.documents_complex(self.schema,
                                                        result_type="ONLY_ID",
                                                        query={
                                                            "and": [{
                                                                        "field": key,
                                                                        "type": "eq",
                                                                        "value":
                                                                            doc_content[
                                                                                key]}
                                                                    for key in
                                                                    doc_content.keys()]
                                                                   + [
                                                                       {
                                                                           "field": "_id",
                                                                           "type": "eq",
                                                                           "value": doc.document_id}
                                                                   ]
                                                        }
                                                        )
            self.assertEquals(
                doc.document_id, str(res.IDs[0])
            )

        self.assertEquals(
            0,
            len(
                self.chino.searches.documents_complex(self.schema,
                                                      result_type="NO_CONTENT",
                                                      query={
                                                          "or": [
                                                              {
                                                                  "field": "fieldInt",
                                                                  "type": "eq",
                                                                  "value": max + 1},
                                                              {"not": [
                                                                  {
                                                                      "field": "fieldInt",
                                                                      "type": "lte",
                                                                      "value": max}
                                                              ]}
                                                          ]
                                                      }
                                                      ).documents
            )
        )

        count_docs = self.chino.searches.documents_complex(self.schema,
                                                           result_type="COUNT",
                                                           query={
                                                               "field": "fieldDate",
                                                               "type": "gte",
                                                               "value": '2018-12-19'}
                                                           )
        self.assertEqual(max, count_docs)

        all_ids = self.chino.searches.documents_complex(self.schema,
                                                        result_type="ONLY_ID",
                                                        query={
                                                            "field": "fieldDate",
                                                            "type": "gte",
                                                            "value": '2018-12-19'}
                                                        )
        self.assertIn(doc.document_id,
                      [str(_id['id']) for _id in all_ids.to_dict()["IDs"]])

        self.assertEquals(
            doc.document_id,  # last created document
            self.chino.searches.documents_complex(self.schema,
                                                  sort=[dict(field="fieldInt",
                                                             order="desc")],
                                                  query={"field": "fieldDate",
                                                         "type": "gte",
                                                         "value": '2018-12-19'},
                                                  limit=1
                                                  ).documents[0].document_id
        )


class SearchUsersChinoTest(BaseChinoTest):
    def setUp(self):
        super(SearchUsersChinoTest, self).setUp()
        fields = [dict(name='fieldInt', type='integer', indexed=True),
                  dict(name='fieldString', type='string', indexed=True),
                  dict(name='fieldBool', type='boolean', indexed=True),
                  dict(name='fieldDate', type='date', indexed=True),
                  dict(name='fieldDateTime', type='datetime', indexed=True)]
        self.schema = self.chino.user_schemas.create('test', fields)._id

    def tearDown(self):
        users = self.chino.users.list(self.schema)
        for user in users.users:
            self.chino.users.delete(user._id, force=True)
        self.chino.user_schemas.delete(self.schema, True)

    # DEPRECATED: to be removed in v>=3.0.0
    def test_search_users(self):
        for i in range(9):
            self.chino.users.create(self.schema, username="user_test_%s" % i,
                                    password='1234567890AAaa',
                                    attributes=dict(fieldInt=123,
                                                    fieldString='test',
                                                    fieldBool=False,
                                                    fieldDate='2015-02-19',
                                                    fieldDateTime='2015-02-19T16:39:47'))
        time.sleep(float(cfg.sleep) / 2.0)
        last_doc = self.chino.users.create(self.schema,
                                           username="user_test_last",
                                           password='1234567890AAaa',
                                           attributes=dict(fieldInt=123,
                                                           fieldString='test',
                                                           fieldBool=False,
                                                           fieldDate='2015-02-19',
                                                           fieldDateTime='2015-02-19T16:39:47'))

        # self.chino.searches.search(self.schema) # TODO: improve tests
        time.sleep(
            float(cfg.sleep) * 3.0)  # wait twice the index max update time

        res = self.chino.searches.users(self.schema, filters=[
            {"field": "fieldInt", "type": "eq", "value": 123}])
        self.assertEqual(res.paging.total_count, 10, res)
        res = self.chino.searches.users(self.schema,
                                        filters=[
                                            {"field": "username", "type": "eq",
                                             "value": 'user_test_last'}])
        self.assertEqual(res.paging.total_count, 1, res)
        res = self.chino.searches.users(self.schema, filters=[
            {"field": "username", "type": "eq", "value": 'user_test_last'}],
                                        result_type="EXISTS")
        self.assertEqual(res, True, res)
        res = self.chino.searches.users(self.schema, filters=[
            {"field": "username", "type": "eq", "value": 'user_test_last'}],
                                        result_type="USERNAME_EXISTS")
        self.assertEqual(res, True, res)
        self.chino.users.delete(last_doc.user_id, force=True)
        time.sleep(float(cfg.sleep))
        res = self.chino.searches.users(self.schema, filters=[
            {"field": "fieldInt", "type": "eq", "value": 123}])
        self.assertEqual(res.paging.total_count, 9, res)

        res = self.chino.searches.users_complex(self.schema,
                                                query={"field": "fieldInt",
                                                       "type": "eq",
                                                       "value": 123})
        self.assertEqual(res.paging.total_count, 9, res)

    # DEPRECATED: to be removed in v>=3.0.0
    def test_search_users_consistent(self):
        doc = None
        for i in range(4):
            doc = self.chino.users.create(self.schema,
                                          username="user_test_%s" % i,
                                          password='1234567890AAaa',
                                          attributes=dict(fieldInt=123,
                                                          fieldString='test',
                                                          fieldBool=False,
                                                          fieldDate='2015-02-19',
                                                          fieldDateTime='2015-02-19T16:39:47'),
                                          consistent=True)
        res = self.chino.searches.users(self.schema, filters=[
            {"field": "fieldInt", "type": "eq", "value": 123}])
        self.assertEqual(res.paging.total_count, i + 1, res)

        # self.chino.users.delete(doc.user_id, force=True,consistent=True)
        # res = self.chino.searches.users(self.schema, filters=[{"field": "fieldInt", "type": "eq", "value": 123}])
        # self.assertEqual(res.paging.total_count, 9, res)

    # DEPRECATED: to be removed in v>=3.0.0
    def test_search_users_complex(self):
        usr = None
        max = 4
        for i in range(max):
            usr_attributes = dict(fieldInt=i, fieldString='test',
                                  fieldBool=False,
                                  fieldDate='2018-12-19',
                                  fieldDateTime='2018-12-19T16:39:47')
            usr = self.chino.users.create(self.schema,
                                          username="user_test_%s" % i,
                                          password='1234567890AAaa',
                                          attributes=usr_attributes,
                                          consistent=True
                                          )
            self.assertTrue(
                self.chino.searches.users_complex(self.schema,
                                                  result_type="EXISTS",
                                                  query={
                                                      "and": [{"field": key,
                                                               "type": "eq",
                                                               "value":
                                                                   usr_attributes[
                                                                       key]}
                                                              for key in
                                                              usr_attributes.keys()]
                                                             + [
                                                                 {
                                                                     "field": "_id",
                                                                     "type": "eq",
                                                                     "value": usr.user_id}
                                                             ]
                                                  }
                                                  )
            )
            self.assertTrue(
                self.chino.searches.users_complex(self.schema,
                                                  result_type="USERNAME_EXISTS",
                                                  query={"field": "username",
                                                         "type": "eq",
                                                         "value": usr.username}
                                                  )
            )

        self.assertFalse(
            self.chino.searches.users_complex(self.schema,
                                              result_type="EXISTS",
                                              query={
                                                  "or": [
                                                      {"field": "fieldInt",
                                                       "type": "eq",
                                                       "value": max + 1},
                                                      {"not": [
                                                          {"field": "fieldInt",
                                                           "type": "lte",
                                                           "value": max}
                                                      ]}
                                                  ]
                                              }
                                              )
        )

        count_docs = self.chino.searches.users_complex(self.schema,
                                                       result_type="COUNT",
                                                       query={
                                                           "field": "fieldDate",
                                                           "type": "gte",
                                                           "value": '2018-12-19'}
                                                       )
        self.assertEqual(max, count_docs)

        self.assertEquals(
            usr.user_id,  # last created user
            self.chino.searches.users_complex(self.schema,
                                              sort=[dict(field="fieldInt",
                                                         order="desc")],
                                              query={"field": "fieldDate",
                                                     "type": "gte",
                                                     "value": '2018-12-19'},
                                              limit=1
                                              ).users[0].user_id
        )


class PermissionChinoTest(BaseChinoTest):
    def setUp(self):
        super(PermissionChinoTest, self).setUp()
        groups = self.chino.groups.list()
        for g in groups.groups:
            self.chino.groups.delete(g._id, True)
        u_schemas = self.chino.user_schemas.list()
        for us in u_schemas.user_schemas:
            li = self.chino.users.list(us._id)
            for user in li.users:
                self.chino.users.delete(user._id, force=True)
        fields = [dict(name='first_name', type='string'),
                  dict(name='last_name', type='string'),
                  dict(name='email', type='string')]
        self.user_schema = self.chino.user_schemas.create('test', fields)._id
        self.password = '12345678'
        self.user0 = self.chino.users.create(user_schema_id=self.user_schema,
                                             username='test',
                                             password=self.password,
                                             attributes=dict(
                                                 first_name='user_0',
                                                 last_name='doe',
                                                 email='test@chino.io'))
        self.user1 = self.chino.users.create(user_schema_id=self.user_schema,
                                             username='test1',
                                             password=self.password,
                                             attributes=dict(
                                                 first_name='user_1',
                                                 last_name='doe',
                                                 email='test@chino.io'))

        self.group = self.chino.groups.create('testing',
                                              attributes=dict(hospital='test'))

        fields = [dict(name='fieldInt', type='integer'),
                  dict(name='fieldString', type='string'),
                  dict(name='fieldBool', type='boolean'),
                  dict(name='fieldDate', type='date'),
                  dict(name='fieldDateTime', type='datetime')]
        self.repo = self.chino.repositories.create('test')._id
        self.schema = self.chino.schemas.create(self.repo, 'test', fields)
        app = self.chino.applications.create("test", grant_type='password')
        self.chino_user0 = ChinoAPIClient(customer_id=cfg.customer_id,
                                          customer_key=cfg.customer_key,
                                          url=cfg.url, client_id=app.app_id,
                                          client_secret=app.app_secret,
                                          force_https=False)
        self.chino_user0.users.login(self.user0.username, self.password)
        self.chino_user1 = ChinoAPIClient(customer_id=cfg.customer_id,
                                          customer_key=cfg.customer_key,
                                          url=cfg.url, client_id=app.app_id,
                                          client_secret=app.app_secret,
                                          force_https=False)
        self.chino_user1.users.login(self.user1.username, self.password)

    def tearDown(self):
        u_schemas = self.chino.user_schemas.list()
        for us in u_schemas.user_schemas:
            li = self.chino.users.list(us._id)
            for user in li.users:
                self.chino.users.delete(user._id, force=True)
            self.chino.user_schemas.delete(us._id, True)
        groups = self.chino.groups.list()
        for g in groups.groups:
            self.chino.groups.delete(g._id, True)
        try:
            self.chino.schemas.delete(self.schema, True, all_content=True)
            self.chino.repositories.delete(self.repo, True, all_content=True)
        except CallError:
            pass

    def test_create_a_repository(self):
        with self.assertRaises(
                CallError):  # user has no create permission, expected an error.
            self.repository = self.chino_user0.repositories.create('test')
        self.chino.permissions.resources('grant', 'repositories', 'users',
                                         self.user0._id,
                                         manage=['C', 'R', 'U', 'L'],
                                         authorize=['A'])
        # now it has the permissions
        self.repository = self.chino_user0.repositories.create('test')
        # grant to user1 the permission of read all the repository but not to create
        self.chino.permissions.resources('grant', 'repositories', 'users',
                                         self.user1._id,
                                         manage=['R', 'U', 'L'])
        with self.assertRaises(
                CallError):  # must fail because user has no create permission
            self.chino_user1.repositories.create('test')
        li = self.chino_user1.repositories.list()
        for repo in li.repositories:
            self.chino_user1.repositories.detail(repo._id)
        with self.assertRaises(
                CallError):  # must fail because user has no "D" permission
            self.chino_user1.repositories.delete(self.repository._id)
        self.chino_user0.permissions.resource('grant', 'repositories',
                                              self.repository._id, 'users',
                                              self.user1._id,
                                              manage=['D'])
        with self.assertRaises(
                CallError):  # must fail because user is not authorized
            self.chino_user1.permissions.resource('grant', 'repositories',
                                                  self.repository._id, 'users',
                                                  self.user0._id,
                                                  manage=['D'])
        self.chino_user1.repositories.delete(self.repository._id)

    def test_create_documents(self):
        self.chino.permissions.resources('grant', 'repositories', 'users',
                                         self.user0._id, manage=['C'])
        self.repository = self.chino_user0.repositories.create('test')
        fields = [dict(name='fieldInt', type='integer'),
                  dict(name='fieldString', type='string'),
                  dict(name='fieldBool', type='boolean'),
                  dict(name='fieldDate', type='date'),
                  dict(name='fieldDateTime', type='datetime')]
        self.schema = self.chino_user0.schemas.create(self.repository._id,
                                                      'test_schema',
                                                      fields)._id
        content = dict(fieldInt=123, fieldString='test', fieldBool=False,
                       fieldDate='2015-02-19',
                       fieldDateTime='2015-02-19T16:39:47')
        self.document = self.chino_user0.documents.create(self.schema,
                                                          content=content)._id
        with self.assertRaises(
                CallError):  # must fail because user is not authorized
            self.chino_user1.documents.create(self.schema, content=content)._id
        with self.assertRaises(
                CallError):  # must fail because user is not authorized
            self.chino_user1.documents.list(self.schema)
        self.chino_user0.permissions.resource_children('grant', 'schemas',
                                                       self.schema,
                                                       'documents', 'users',
                                                       self.user1._id,
                                                       manage=['L'])
        self.assertEqual(
            self.chino_user1.documents.list(self.schema).paging.total_count, 0)
        self.assertEqual(
            self.chino_user0.documents.list(self.schema).paging.total_count, 1)


# # NOTE: this test case will be removed along with the Consent Management API
#         in the upcoming major version (3.0.0)
# @unittest.skipIf(cfg.url.startswith('https://api.chino.io'),
#                  "not on production")
# @unittest.skipIf(os.environ.get('SKIP_PROD', False),
#                  "Not for production like systems")
# class ConsentChinoTest(BaseChinoTest):
#
#     def setUp(self):
#         super(ConsentChinoTest, self).setUp()
#         self.details = {
#             "description": "This policy was created for test purposes. The policy url is linked to Chino.io policy.",
#             "policy_url": "https://www.chino.io/legal/privacy-policy",
#             "policy_version": "test",
#             "collection_mode": "none"
#         }
#         self.data_controller = {
#             "company": "Chino.io",
#             "contact": "controller",
#             "address": "Via S.G. Bosco 27, 38068 Rovereto",
#             "email": "controller@mail.tld",
#             "VAT": "n/d",
#             "on_behalf": True
#         }
#         self.purposes = [
#             {
#                 "authorized": True,
#                 "purpose": "testing",
#                 "description": "Testing class api.ChinoAPIConsents"
#             },
#             {
#                 "authorized": False,
#                 "purpose": "testing",
#                 "description": "Testing class objects.Consent"
#             }
#         ]
#         self.user_id = "user_id"
#         self.user_id_alt = "another-user_id"
#         # list of Consents for test_list and test_withdraw
#         self.consent_ls = []
#         self.consent_ls.append(
#             self.chino.consents.create(self.user_id, self.details,
#                                        self.data_controller, self.purposes))
#         self.consent_ls.append(
#             self.chino.consents.create(self.user_id_alt, self.details,
#                                        self.data_controller, self.purposes))
#         self.consent_ls.append(
#             self.chino.consents.create(self.user_id, self.details,
#                                        self.data_controller, self.purposes))
#
#     def tearDown(self):
#         ls = self.chino.consents.list(limit=100)
#         for c in ls.consents:
#             if c.to_dict()['policy_version'] == 'test':
#                 self.chino.consents.delete(c._id)
#
#     def test_list(self):
#         ls = self.chino.consents.list()
#         self.assertIsNotNone(ls.paging)
#         self.assertIsNotNone(ls.consents)
#
#         ls = self.chino.consents.list(self.user_id)
#         for consent in ls.consents:
#             self.assertEqual(consent.to_dict()['user_id'], self.user_id)
#
#     def test_CRUD(self):
#         USER_ID1 = "username@mail.tld"
#         USER_ID2 = "another_username@mail.tld"
#         # CREATE
#         new = self.chino.consents.create(USER_ID1, self.details,
#                                          self.data_controller, self.purposes)
#         list = self.chino.consents.list(limit=100)
#         self.assertGreater(list.paging.count, 0)
#         read_id = 0
#         for consent in list.consents:
#             if consent._id == new._id:
#                 read_id = consent._id
#         self.assertFalse(read_id == 0)
#         # READ
#         read = self.chino.consents.detail(read_id)
#         read_compare = self.chino.consents.detail(new._id)
#         self.assertTrue(self._equals(new.to_dict(), read.to_dict()),
#                         msg="created:\n%s\n\nread:\n\n%s\n" % (
#                         new.to_json(), read.to_json())
#                         )
#         # UPDATE
#         updated = self.chino.consents.update(read._id, USER_ID2, self.details,
#                                              self.data_controller,
#                                              self.purposes)
#         self.assertFalse(self._equals(new.to_dict(), updated.to_dict()),
#                          msg=(
#                                      "Update returned an unmodified object.\nid1: %s\nid2: %s" % (
#                              new._id, updated._id)))
#
#         read = self.chino.consents.detail(read_id)
#         self.assertTrue(self._equals(updated.to_dict(), read.to_dict()),
#                         msg="updated:\n%s\n\nread:\n\n%s\n" % (
#                         updated.to_json(), read.to_json())
#                         )
#         history = self.chino.consents.history(read._id)
#         self.assertGreater(len(history.consents), 1)
#         # DELETE
#         self.chino.consents.delete(read._id)
#         for cid in [new._id, read._id, updated._id]:
#             self.assertRaises(CallError,
#                               self.chino.consents.detail,
#                               {'consent_id': cid}
#                               )
#
#     def test_withdraw(self):
#         withdraw = self.chino.consents.list(user_id=self.user_id_alt).consents[
#             0]
#         self.chino.consents.withdraw(withdraw._id)
#
#         read = self.chino.consents.detail(withdraw._id)
#         self.assertIsNotNone(read.to_dict()['withdrawn_date'])


if __name__ == '__main__':
    unittest.main()



__author__ = 'Stefano Tranquillini <stefano@chino.io>'

from setuptools import setup

setup(
    setup_requires=['pbr'],
    pbr=True,
)
# # Always prefer setuptools over distutils
# from setuptools import setup, find_packages
# # To use a consistent encoding
# from codecs import open
# from os import path
# here = path.abspath(path.dirname(__file__))
# # Utility function to read the README file.
# # Used for the long_description.  It's nice, because now 1) we have a top level
# # README file and 2) it's easier to type in the README file than to put a raw
# # string in below ...
# def read(fname):
#     return open(path.join(path.dirname(__file__), fname)).read()
#
# with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
#     long_description = f.read()
# setup(name='chino',
#       version='1.0.10',
#       description='Wrapper for Chino.io API',
#       long_description=long_description,
#       author='Stefano Tranquillini',
#       author_email='stefano@chino.io',
#       url='https://www.chino.io',
#       packages=['chino'],
#       package_data={
#           "chino": [
#               "logging.conf"
#           ]},
#       license = 'CC BY-SA 4.0',
#
#       classifiers=[
#            "Topic :: Software Development",
#            "Development Status :: 5 - Production/Stable",
#            "Programming Language :: Python :: 2",
#            "Programming Language :: Python :: 3",
#           ],
#       )



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import json
from collections import namedtuple

__author__ = 'Stefano Tranquillini <stefano@chino.io>'

import csv
import hashlib
import io
import logging

logger = logging.getLogger('chino')


class ChinoBaseObject(object):
    __str_name__ = 'not set'
    __str_names__ = 'not set'
    exclude_repr_field = ['_id', 'class_obj']
    only_key_field = []
    @property
    def _id(self, indent=None):
        return '-'

    # def to_json(self, indent=None):
    #     # raise NotImplementedError()
    #     return json.dumps(self, default=lambda o: o.to_dict() if hasattr(o,'to_dict()') else o.__dict__,
    #                       sort_keys=True, indent=indent)

    def to_dict(self):
        # make a copy here
        res = dict()
        for k, v in vars(self).items():
            if k not in ['_id', 'class_obj']:
                try:
                    res[k] = v.to_dict()
                except:
                    res[k] = v
        return res

    def to_json(self, indent=None, sort_keys=True):
        return json.dumps(self.to_dict(), indent=indent, sort_keys=sort_keys)

    def __repr__(self):
        """
        This returns the class name and parameters
        :return:
        """
        s = "%s(" % ".".join(
            [self.__class__.__module__, self.__class__.__name__])
        res = []
        for k, v in vars(self).items():
            if k not in self.exclude_repr_field:
                if k in self.only_key_field:
                    res.append("%s=%s" % (k, "..."))
                else:
                    res.append("%s=%s" % (k, v.__repr__()))
        s += ",".join(res)
        s += ")"
        return s
        # return self.to_dict*


class Paging(ChinoBaseObject):

    def __init__(self, offset=0, limit=100, count=-1, total_count=-1,
                 **kwargs):
        super(Paging, self).__init__()
        self.offset = offset
        self.limit = limit
        self.total_count = total_count
        self.count = count

    def as_map(self):
        return {
            'offset': self.offset,
            'limit': self.limit,
        }

    # def __str__(self):
    #     return "Paging [offset:%s,limit:%s,count:%s,total_count:%s]" % (self.offset, self.limit, self.count, self.total_count)


class ListResult(ChinoBaseObject):

    def __init__(self, class_obj, result, **kwargs):
        self.paging = Paging(result['offset'], result['limit'], result[
            'count'], result['total_count'])
        results = []
        for r in result[class_obj.__str_names__]:
            if type(r) == dict:
                results.append(class_obj(**r))
            else:
                results.append(class_obj(r))
        self.__setattr__(class_obj.__str_names__, results)
        self.class_obj = class_obj.__str_names__

    def to_dict(self):
        res = super(ListResult, self).to_dict()
        res['paging'] = self.paging.to_dict()
        res[self.class_obj] = [o.to_dict() for o in
                               self.__getattribute__(self.class_obj)]
        return res

    def __len__(self):
        return self.paging.total_count

    def __getitem__(self, item):
        return self.__getattribute__(self.class_obj)[item]

    def __iter__(self):
        return iter(self.__getattribute__(self.class_obj))


class _DictContent(ChinoBaseObject):
    """
    Subclass for the content that is a dict.

    Example::

        "content": {
            "physician_id": "130b875a-d291-453e-8559-1b3c54500432",
            "physician_name": "Jack",
            "observation": "The patient was ok.",
            "visit_date": "2015-02-19 16:39:47"
        }
    """

    def __init__(self, *args, **kwargs):
        for k, v in kwargs.items():
            self.__setattr__(k, v)

    @property
    def _id(self):
        return "-"


class IDs(ChinoBaseObject):
    __str_name__ = 'ID'
    __str_names__ = 'IDs'

    def __init__(self, id):
        self.id = id

    def __repr__(self):
        return self.id

    def __str__(self):
        return self.id


class Repository(ChinoBaseObject):
    """

    Example::

            "repository": {
              "repository_id": "d88084ef-b6f7-405d-9863-d35b99543389",
              "last_update": "2015-02-24T21:48:16.332",
              "is_active": true,
              "description": "This is a test repo",
              "insert_date": "2015-02-24T21:48:16.332"
            }

    """
    __str_name__ = 'repository'
    __str_names__ = 'repositories'

    @property
    def _id(self):
        return self.repository_id

    def __init__(self, repository_id=None, insert_date=None, last_update=None,
                 description=None, is_active=True, **kwargs):
        self.repository_id = repository_id
        self.description = description
        self.insert_date = insert_date
        self.last_update = last_update
        self.is_active = is_active


class User(ChinoBaseObject):
    """

    Example::

        "user": {
          "username": "james",
          "user_id": "d88084ef-b6f7-405d-9863-d35b99543389",
          "insert_date": "2015-02-05T10:53:38.157",
          "last_update": "2015-02-05T10:53:38.157",
          "is_active": true,
          "attributes": {
            "first_name": "James",
            "last_name": "Ford",
            "email": "james@acme.com"
          },
          "groups": [
            "d88084ef-b6f7-405d-9863-d35b99543389",
            "1eb39c88-3ac8-4664-b897-849dd260c72b"
          ]
        }

    For the creation it has `password`
    """
    __str_name__ = 'user'
    __str_names__ = 'users'

    @property
    def _id(self):
        return self.user_id

    def to_dict(self):
        res = super(User, self).to_dict()
        if hasattr(res, 'password'):
            del res['password']
        if self.attributes:
            if type(self.attributes) is not dict:
                res['attributes'] = self.attributes.to_dict()
        return res

    def __init__(self, user_id=None, username=None, insert_date=None,
                 schema_id=None, last_update=None, is_active=None,
                 attributes=None,
                 groups=None, password=None, **kwargs):
        self.username = username
        self.user_id = user_id
        self.insert_date = insert_date
        self.last_update = last_update
        self.schema_id = schema_id
        self.is_active = is_active
        self.attributes = _DictContent(**attributes) if attributes else None
        self.groups = groups
        self.password = password


# TODO: add attribute funciton

class Group(ChinoBaseObject):
    """

    Example::

        "group": {
          "groupname": "pysiciansXY",
          "group_id": "77de01d8-492d-4cc2-a2b2-d3e76edc0657",
          "insert_date": "2015-02-07T12:14:46.754",
          "is_active": true,
          "last_update": "2015-02-21T19:07:45.832",
          "attributes": {
            "hospital": "Main Hospital"
          }
        }

    """

    __str_name__ = 'group'
    __str_names__ = 'groups'

    @property
    def _id(self):
        return self.group_id

    def to_dict(self):
        res = super(Group, self).to_dict()
        if self.attributes:
            if type(self.attributes) is not dict:
                res['attributes'] = self.attributes.to_dict()
        return res

    def __init__(self, group_id=None, group_name=None, insert_date=None,
                 is_active=None, last_update=None,
                 attributes=None, **kwargs):
        """Constructor for Group"""
        self.group_id = group_id
        self.group_name = group_name
        self.insert_date = insert_date
        self.last_update = last_update
        self.is_active = is_active
        self.attributes = _DictContent(**attributes) if attributes else None


class Document(ChinoBaseObject):
    """
    Class for the document.

    Example::

        "document": {
          "document_id": "c373ba1a-1f23-4e36-a099-0ea3306b093d",
          "repository_id": "3ddba8af-6965-4416-9c5c-acf6af95539d",
          "schema_id": "130b875a-d291-453e-8559-1b3c54500432",
          "insert_date": "2015-02-24T22:27:35.919",
          "is_active": true,
          "last_update": "2015-02-24T22:27:35.919",
          "content": {
            "physician_id": "130b875a-d291-453e-8559-1b3c54500432",
            "physician_name": "Jack",
            "observation": "The patient was ok.",
            "visit_date": "2015-02-19 16:39:47"
          }
        }
    """
    __str_name__ = 'document'
    __str_names__ = 'documents'

    @property
    def _id(self):
        return self.document_id

    def to_dict(self):
        res = super(Document, self).to_dict()
        if self.content:
            res['content'] = self.content.to_dict()
        return res

    def __init__(self, document_id=None, repository_id=None, schema_id=None,
                 insert_date=None, last_update=True,
                 content=None, is_active=False, **kwargs):
        self.document_id = document_id
        self.repository_id = repository_id
        self.schema_id = schema_id
        self.insert_date = insert_date
        self.last_update = last_update
        self.is_active = is_active
        self.content = _DictContent(**content) if content else None


class _Field(ChinoBaseObject):

    def __init__(self, type, name, indexed=None, **kwargs):
        self.type = type
        self.name = name
        if indexed:
            self.indexed = indexed

        # like this we manage boh an unset default and a default set to None
        if 'default' in kwargs:
            self.default = kwargs.get('default')

    __str_name__ = 'field'
    __str_names__ = 'fields'

    @property
    def _id(self):
        return '-'


class _Fields(ChinoBaseObject):

    def __init__(self, fields):
        self.fields = fields


class Schema(ChinoBaseObject):
    """

    Example::

            "schema": {
              "description": "testSchema",
              "schema_id": "b1cc4a53-19a1-4819-a8c7-20bf153ec9cf",
              "repository_id": "3ddba8af-6965-4416-9c5c-acf6af95539d",
              "is_active": true,
              "insert_date": "2015-02-15T16:22:04.058",
              "last_update": "2015-02-15T16:22:04.058",
              "structure": {
                "fields": [
                  {
                    "type": "integer",
                    "name": "test_int"
                  },
                  {
                    "type": "float",
                    "name": "test_float"
                  },
                  {
                    "type": "string",
                    "name": "test_string"
                  },
                  {
                    "type": "boolean",
                    "name": "test_boolean"
                  },
                  {
                    "type": "data",
                    "name": "test_date"
                  },
                  {
                    "type": "time",
                    "name": "test_time"
                  },
                  {
                    "type": "datetime",
                    "name": "test_datetime"
                  }
                ]
              }
            }
    """

    __str_name__ = 'schema'
    __str_names__ = 'schemas'

    def to_dict(self):
        res = super(Schema, self).to_dict()
        res['structure'] = dict(fields=[f.to_dict()
                                        for f in self.structure.fields])
        return res

    # # print res
    #     if self.structure:
    #         if type(self.structure) is not dict:
    #             res['structure'] = dict(fields=[f.to_dict() for f in self.structure.fields])
    #     return res

    @property
    def _id(self):
        return self.schema_id

    def __init__(self, schema_id=None, description=None, repository_id=None,
                 is_active=None, insert_date=None,
                 last_update=None, structure=None, **kwargs):
        self.schema_id = schema_id
        self.description = description
        self.repository_id = repository_id
        self.is_active = is_active
        self.insert_date = insert_date
        self.last_update = last_update
        # assuming it's a dict.
        if structure:
            self.structure = _Fields(
                fields=[_Field(**f) for f in structure['fields']])


class UserSchema(ChinoBaseObject):
    """
    Similar to Schema
    """
    __str_name__ = 'user_schema'
    __str_names__ = 'user_schemas'

    def to_dict(self):
        res = super(UserSchema, self).to_dict()
        res['structure'] = dict(fields=[f.to_dict()
                                        for f in self.structure.fields])
        return res

    @property
    def _id(self):
        return self.user_schema_id

    def __init__(self, user_schema_id=None, description=None, is_active=None,
                 insert_date=None,
                 last_update=None, structure=None, groups=None, **kwargs):
        self.user_schema_id = user_schema_id
        self.description = description
        self.groups = groups
        self.is_active = is_active
        self.insert_date = insert_date
        self.last_update = last_update
        # assuming it's a dict.
        if structure:
            self.structure = _Fields(
                fields=[_Field(**f) for f in structure['fields']])


class Collection(ChinoBaseObject):
    """
    Class for the Collection.

    Example::

        "collection":
            {
              "collection_id": "de991a83-da39-4a72-a79e-1376124ebd57",
              "last_update": "2016-02-08T11:05:26.571Z",
              "is_active": true,
              "name": "testCollection",
              "insert_date": "2016-02-08T11:05:26.571Z"
            }
    """
    __str_name__ = 'collection'
    __str_names__ = 'collections'

    @property
    def _id(self):
        return self.collection_id

    def __init__(self, collection_id=None, name=None, insert_date=None,
                 last_update=True,
                 content=None, is_active=False, **kwargs):
        self.collection_id = collection_id
        self.name = name
        self.insert_date = insert_date
        self.last_update = last_update
        self.is_active = is_active


class Application(ChinoBaseObject):
    """
    Class for the Collection.

    Example::
          "data": {
            "app_secret": "bBRdHHPmIHbEzywlrK3v9sjbvDXLG20xNhMuyX7g1hAbP4ht7aznD7O7gcKlfhWUZ8osgJezZGK4EainErX4ktxkMX56KtihFiKVtfw8pMVDIVx1N3XjhqTbTTIGv4g7",
            "grant_type": "authorization-code",
            "app_name": "Teest",
            "redirect_url": "http://127.0.0.1/",
            "app_id": "4ke1mor5GW4YtH80Y9eIaAFLHUAwLtQ1l7wOQnQV"
          },
    """
    __str_name__ = 'application'
    __str_names__ = 'applications'

    @property
    def _id(self):
        return self.app_id

    def __init__(self, app_name=None, app_secret=None, app_id=None,
                 redirect_url=None, grant_type=None, **kwargs):
        self.app_name = app_name
        self.app_secret = app_secret
        self.app_id = app_id
        self.redirect_url = redirect_url
        self.grant_type = grant_type
        # self.client_type = client_type


class Permission(ChinoBaseObject):
    """
    Class for the Permission.

    Example::

        {
           "message":null,
           "data":{
              "permissions":[
                 {
                    "access":"Structure",
                    "permission":{
                       "Authorize":[
                          "R",
                          "U",
                          "D",
                          "A"
                       ],
                       "Manage":[
                          "R",
                          "U",
                          "D"
                       ]
                    },
                    "resource_type":"Schema",
                    "resource_id":"6cc90e37-4018-4503-8e9b-c0f3057abd6d"
                 }
              ]
           },
           "result":"success",
           "result_code":200
        }
    """
    __str_name__ = 'permission'
    __str_names__ = 'permissions'

    @property
    def _id(self):
        return "-"

    def __init__(self, access=None, permission=None, resource_type=None,
                 resource_id=None, parent_id=None, **kwargs):
        self.access = access
        if permission:
            self.permission = _PermissionField(**permission)
        self.resource_type = resource_type
        if resource_id:
            self.resource_id = resource_id
        else:
            # TODO: can parent_id be None?
            self.parent_id = parent_id

    def to_dict(self):
        res = super(Permission, self).to_dict()
        res['permission'] = self.permission.to_dict()
        return res


class _PermissionField(ChinoBaseObject):

    def __init__(self, Authorize=None, Manage=None, created_document=None,
                 **kwargs):
        if Authorize:
            self.authorize = Authorize
        if Manage:
            self.manage = Manage
        if created_document:
            self.created_document = created_document

    # TODO: support functions to help in managing grants

    @property
    def _id(self):
        return "-"

    def to_dict(self):
        res = dict()
        if hasattr(self, 'authorize'):
            res['authorize'] = self.authorize
        if hasattr(self, 'manage'):
            res['manage'] = self.manage
        if hasattr(self, 'created_document'):
            res['created_document'] = self.created_document
        return res


class _SortField(ChinoBaseObject):

    def __init__(self, field, order='asc', **kwargs):
        self.field = field
        self.order = order

    @property
    def _id(self):
        return "-"


class _FilterField(ChinoBaseObject):

    def __init__(self, field, value, type='eq', case_sensitive=False,
                 **kwargs):
        self.field = field
        self.type = type
        self.case_sensitive = case_sensitive
        self.value = value

    @property
    def _id(self):
        return "-"


class Search(ChinoBaseObject):

    def __init__(self, schema_id, result_type, sort=None, filters=None,
                 **kwargs):
        self.schema_id = schema_id
        self.result_type = result_type
        self.sort = []
        self.filters = []
        if sort:
            for s in sort:
                if type(s) is dict:
                    self.sort.append(_SortField(**s))
                else:
                    self.sort.append(s)
        if filters:
            for f in filters:
                if type(f) is dict:
                    self.filters.append(_FilterField(**f))
                else:
                    self.sort.append(f)

    def to_dict(self):
        res = super(Search, self).to_dict()
        if self.sort:
            res['sort'] = [s.to_dict() for s in self.sort]
        if self.filters:
            res['filters'] = [f.to_dict() for f in self.filters]
        return res


class Consent(ChinoBaseObject):
    """

    Example:

            consent : {
              "message": null,
              "data": {
                "count": 2,
                "total_count": 2,
                "consents": [
                  {
                    "user_id": "admin@chino.io",
                    "description": "a long textual description",
                    "data_controller": {
                      "company": "Acme",
                      "contact": "John Doe",
                      "address": "221B Baker St.",
                      "email": "info@acme.com",
                      "VAT": "IT03256920228",
                      "on_behalf": true
                    },
                    "consent_id": "0a63b938-c481-42aa-8da2-1db34f2d0dea",
                    "purposes": [
                      {
                        "authorized": true,
                        "purpose": "marketing",
                        "description": "mailing list"
                      }
                    ],
                    "policy_url": "https://chino.io/legal/privacy-policy",
                    "policy_version": "v1.0",
                    "withdrawn_date": null,
                    "inserted_date": "2018-01-26T13:16:28.680Z",
                    "collection_mode": "webform"
                  }
                ],
                "limit": 10,
                "offset": 0
              },
              "result": "success",
              "result_code": 200
            }

    """
    __str_name__ = 'consent'
    __str_names__ = 'consents'

    @property
    def _id(self):
        return self.consent_id

    def to_dict(self):
        res = super(Consent, self).to_dict()
        if self.data_controller:
            res['data_controller'] = self.data_controller.to_dict()
        if self.purposes:
            res['purposes'] = [purpose.to_dict() for purpose in self.purposes]

        return res

    def __init__(self, consent_id=None, user_id=None, description=None,
                 data_controller=None, purposes=None,
                 policy_url=None, policy_version=None, collection_mode=None,
                 inserted_date=None, withdrawn_date=None):
        self.user_id = user_id
        self.description = description
        self.data_controller = _DictContent(
            **data_controller) if data_controller else None
        self.consent_id = consent_id
        # 'purposes' is a list of dict-s
        if purposes:
            self.purposes = []
            for p in purposes:
                self.purposes.append(_DictContent(**p))
        else:
            self.purposes = None

        self.policy_url = policy_url
        self.policy_version = policy_version
        # should be always 'None' when a new instance is created
        self.withdrawn_date = withdrawn_date
        self.inserted_date = inserted_date  # set by server
        self.collection_mode = collection_mode

    def is_withdrawn(self):
        """
        Check if this Consent has been withdrawn.

        :return: True if the 'withdrawn_date' attribute of this instance has been set, otherwise False
        """
        return (self.withdrawn_date != None)


class Blob(ChinoBaseObject):
    only_key_field = 'content'
    @property
    def _id(self):
        return self.blob_id

    def __init__(self, filename=None, content=None, blob_id=None, **kwargs):
        self.filename = filename
        self.content = content
        self.blob_id = blob_id
        # if 'content' not in self.exclude_repr_field:
        #     self.exclude_repr_field.append('content')


class DumpFile(ChinoBaseObject):

    def __init__(self, dump_id, data, status_info=None):
        self.dump_id = dump_id
        self.content = data

        # fill the info
        status_info = status_info or {}
        self.start_date = status_info.get('start_date')
        self.last_update = status_info.get('last_update')
        self.status = status_info.get('status')
        self.records_count = status_info.get('records_count')
        self.size = status_info.get('size')
        self.sha256 = status_info.get('sha256')

        self._json = None
        self._csv = None

    def hash_matches(self):
        """
        Return true if hash computed on the content matches
        self.sha256 (populated by calling dump_status)
        """
        return hashlib.sha256(self.content).hexdigest() == self.sha256

    @property
    def json(self):
        """
        Read the content of the file as json

        :return: list of dicts

        """
        if not self._json:
            self._json = [json.loads(i)
                          for i in self.content.decode().split('\n') if i]
        return self._json

    @property
    def csv(self):
        """
        Read the content of the file as csv

        :return: list of lists

        """
        if not self._csv:
            data = io.StringIO(self.content.decode())
            reader = csv.reader(data, delimiter=';', quotechar='"')
            self._csv = [row for row in reader]
        return self._csv

# Blob = namedtuple('Blob', ['filename', 'content'])
BlobDetail = namedtuple(
    'BlobDetail', ['bytes', 'blob_id', 'sha1', 'document_id', 'md5'])

_PermissionProperty = namedtuple(
    '_PermissionProperty', ['read', 'delete', 'update'])

# class _PermissionProperty(ChinoBaseObject):
# """
#         "read": true,
#         "update": false,
#         "delete": true
#
#     """
#
#     def __init__(self, read=False, update=False, delete=False):
#         self.read = read
#         self.delete = delete
#         self.update = update

# class _Permission(ChinoBaseObject):
#     """
#
#     "permission": {
#                   "insert": true,
#                   "all_data": {
#                     "read": true,
#                     "update": false,
#                     "delete": true
#                   },
#                   "own_data": {
#                     "read": true,
#                     "update": true,
#                     "delete": false
#                   }
#                 }
#     """
#
#     def __init__(self, insert, all_data, own_data):
#         self.insert = insert
#         if type(all_data) is dict:
#             self.all_data = _PermissionProperty(**all_data)
#         else:
#             self.all_data = all_data
#         if type(own_data) is dict:
#             self.own_data = _PermissionProperty(**own_data)
#         else:
#             self.own_data = own_data
#
#     def to_dict(self):
#         res = super(_Permission, self).to_dict()
#         if self.all_data:
#             res['all_data'] = [p.to_dict() for p in self.all_data]
#         if self.own_data:
#             res['own_data'] = [p.to_dict() for p in self.own_data]
#         return res
#
#
# class Permission(ChinoBaseObject):
#     """
#
#     Example::
#
#             "permissions": [
#               {
#                 "repository_id": "3ddba8af-6965-4416-9c5c-acf6af95539d",
#                 "schema_id": "b1cc4a53-19a1-4819-a8c7-20bf153ec9cf",
#                 "group_id": "3ddba8af-6965-4416-9c5c-acf6af95539d",
#                 "user_id": "b1cc4a53-19a1-4819-a8c7-20bf153ec9cf",
#                 "permission": {
#                   "insert": true,
#                   "all_data": {
#                     "read": true,
#                     "update": false,
#                     "delete": true
#                   },
#                   "own_data": {
#                     "read": true,
#                     "update": true,
#                     "delete": false
#                   }
#                 }
#               }
#             ]
#     """
#
#     def __init__(self, repository_id=None, schema_id=None, group_id=None, user_id=None, permission=None):
#         self.repository_id = repository_id
#         self.schema_id = schema_id
#         self.group_id = group_id
#         self.user_id = user_id
#         if permission:
#             if dict(permission) is dict:
#                 self.permission = _Permission(permission)
#             else:
#                 self.permission = permission
#
#     def to_dict(self):
#         res = super(Permission, self).to_dict()
#         if self.permission:
#             res['permission'] = [p.to_dict() for p in self.permission]
#         return res
#
